/*
 * Maildir driver for Pine4.64
 * 
 * Written by Eduardo Chappa <chappa@math.washington.edu>
 * Last Update: January 28, 2006
 *
 * The IMAP toolkit provided in this Distribution is
 * Copyright 2004 University of Washington.
 * The full text of our legal notices is contained in the file called
 * CPYRIGHT, included with this Distribution.
 */

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
extern int errno;		/* just in case */
#include "mail.h"
#include "osdep.h"
#include <pwd.h>
#include <sys/stat.h>
#include <sys/time.h>
#include "maildir.h"
#include "rfc822.h"
#include "fdstring.h"
#include "misc.h"
#include "dummy.h"

/* Driver dispatch used by MAIL */
DRIVER maildirdriver = {
  "md",				/* driver name, yes it's md, not maildir */
				/* driver flags */
  DR_MAIL|DR_LOCAL|DR_NAMESPACE|DR_NOSTICKY,
  (DRIVER *) NIL,		/* next driver 				*/
  maildir_valid,		/* mailbox is valid for us 		*/
  maildir_parameters,		/* manipulate parameters		*/
  NIL,				/* scan mailboxes 			*/
  maildir_list,			/* find mailboxes 			*/
  maildir_lsub,			/* find subscribed mailboxes 		*/
  maildir_sub,			/* subscribe to mailbox 		*/
  maildir_unsub,		/* unsubscribe from mailbox 		*/
  maildir_create,		/* create mailbox 			*/
  maildir_delete,		/* delete mailbox 			*/
  maildir_rename,		/* rename mailbox 			*/
  mail_status_default,		/* status of mailbox 			*/
  maildir_open,			/* open mailbox				*/
  maildir_close,		/* close mailbox 			*/
  maildir_fast,			/* fetch message "fast" attributes	*/
  NIL,				/* fetch message flags 			*/
  NIL,				/* fetch overview 			*/
  NIL,				/* fetch message structure 		*/
  maildir_header,		/* fetch message header 		*/
  maildir_text,			/* fetch message body 			*/
  NIL,				/* fetch partial message text 		*/
  NIL,				/* unique identifier 			*/
  NIL,				/* message number 			*/
  NIL,				/* modify flags 			*/
  maildir_flagmsg,		/* per-message modify flags 		*/
  NIL,				/* search for message based on criteria */
  NIL,				/* sort messages 			*/
  NIL,				/* thread messages 			*/
  maildir_ping,			/* ping mailbox to see if still alive 	*/
  maildir_check,		/* check for new messages		*/
  maildir_expunge,		/* expunge deleted messages 		*/
  maildir_copy,			/* copy messages to another mailbox 	*/
  maildir_append,		/* append string message to mailbox 	*/
  NIL				/* garbage collect stream 		*/
};


DRIVER courierdriver = {
  "mc",				/* Why a separate driver? So that
				   createproto will work		*/
				/* driver flags */
  DR_MAIL|DR_LOCAL|DR_NAMESPACE|DR_NOSTICKY,
  (DRIVER *) NIL,		/* next driver 				*/
  maildir_valid,		/* mailbox is valid for us 		*/
  maildir_parameters,		/* manipulate parameters		*/
  NIL,				/* scan mailboxes 			*/
  courier_list,			/* find mailboxes 			*/
  maildir_lsub,			/* find subscribed mailboxes 		*/
  maildir_sub,			/* subscribe to mailbox 		*/
  maildir_unsub,		/* unsubscribe from mailbox 		*/
  maildir_create,		/* create mailbox 			*/
  maildir_delete,		/* delete mailbox 			*/
  maildir_rename,		/* rename mailbox 			*/
  mail_status_default,		/* status of mailbox 			*/
  maildir_open,			/* open mailbox				*/
  maildir_close,		/* close mailbox 			*/
  maildir_fast,			/* fetch message "fast" attributes	*/
  NIL,				/* fetch message flags 			*/
  NIL,				/* fetch overview 			*/
  NIL,				/* fetch message structure 		*/
  maildir_header,		/* fetch message header 		*/
  maildir_text,			/* fetch message body 			*/
  NIL,				/* fetch partial message text 		*/
  NIL,				/* unique identifier 			*/
  NIL,				/* message number 			*/
  NIL,				/* modify flags 			*/
  maildir_flagmsg,		/* per-message modify flags 		*/
  NIL,				/* search for message based on criteria */
  NIL,				/* sort messages 			*/
  NIL,				/* thread messages 			*/
  maildir_ping,			/* ping mailbox to see if still alive 	*/
  maildir_check,		/* check for new messages		*/
  maildir_expunge,		/* expunge deleted messages 		*/
  maildir_copy,			/* copy messages to another mailbox 	*/
  maildir_append,		/* append string message to mailbox 	*/
  NIL				/* garbage collect stream 		*/
};

MAILSTREAM maildirproto = {&maildirdriver};	/* prototype stream */
MAILSTREAM courierproto = {&courierdriver};	/* prototype stream */

void 
md_domain_name(void)
{
   int i;

   strcpy(mdlocaldomain,mylocalhost ());
   for (i = 0; mdlocaldomain[i] ; i++)
      if(mdlocaldomain[i] == '/')
	 mdlocaldomain[i] = '\057';
      else if (mdlocaldomain[i] == ':')
	 mdlocaldomain[i] =  '\072';
}

char *
myrootdir(char *name)
{
return myhomedir();
}

char *
mdirpath(void)
{
  char *path = maildir_parameters(GET_INBOXPATH,NIL);
  return path ? (*path ? path : ".") : "Maildir";
}

/* remove the "#md/" or "#mc/" part from a folder name */
void maildir_remove_root (char **name)
{
  int courier = IS_COURIER(*name);
  char realname[MAILTMPLEN], *p;

  if (maildir_valid_name(*name)){
     (*name) += 3;
     if (**name == '/')
	(*name)++;
  }
  if(courier)
     courier_realname(*name, realname);
  else
     strcpy(realname, *name);
  *name = cpystr(realname);
}


/* Check validity of the name, we accept:
 *	a) #md/directory/folder
 *	b) #md/inbox
 * A few considerations: We can only accept as valid
 *  a) names that start with #md/ and the directory exists or
 *  b) names that do not start with #md/ but are maildir directories (have
 *     the /cur, /tmp and /new structure)
 */
int maildir_valid_name (char *name)
{
  char tmpname[MAILTMPLEN] = {'\0'};

   if (mdfpath)
      fs_give((void **)&mdfpath);
   if (name && (name[0] != '#'))
	sprintf(tmpname,"%s%s",MDPREFIX(CCLIENT), name);
   mdfpath = cpystr(tmpname[0] ? tmpname : name);

  return IS_CCLIENT(name) || IS_COURIER(name);
}

/* Check if the directory whose path is given by name is a valid maildir
 *  directory (contains /cur, /tmp and /new)
 */
int maildir_valid_dir (char *name)
{
 int len;
 DirNamesType i;
 struct stat sbuf;
 char tmp[MAILTMPLEN];

   if(name[strlen(name) - 1] == '/')
      name[strlen(name) - 1] = '\0';
   len = strlen(name);
   for (i = Cur; i != EndDir; i++){
      MDFLD(tmp, name, i);
      if (stat(tmp, &sbuf) < 0 || !S_ISDIR(sbuf.st_mode))
	  break;
   }
   name[len] = '\0';
   return (i == EndDir) ? T : NIL;
}

void courier_realname(char *name, char *realname)
{
  int i,j;

  if(!name)
    return;

  for (i = 0, j = 0; i < MAILTMPLEN && j < strlen(name); j++, i++){
      realname[i] = name[j];
      if(name[j] == '/' && name[j+1] != '.' && name[j+1] != '%' 
		&& name[j+1] != '*') 
	realname[++i] = '.';
  }
  if(realname[i-1] == '.')
    i--; 
  realname[i] = '\0';
}


/* given a maildir folder, return its path. Memory freed by caller. Directory
 * does not contain the trailing slash "/". On error NULL is returned.
 */
int maildir_file_path (char *name, char *tmp)
{
   char *maildirpath = mdirpath();
   char realname[MAILTMPLEN];
   int courier = IS_COURIER(name);

   /* There are several ways in which the path can come, so we will handle 
      them here. First we deal with #mc/ or #md/ prefix by removing the 
      prefix, if any */

    maildir_remove_root(&name);
    tmp[0] = '\0';	/* just in case something fails */

    if (strlen(myrootdir(name)) + 
		max(strlen(name), strlen(maildirpath)) > MAILTMPLEN){
	errno = ENAMETOOLONG;
	sprintf(tmp,"Error opening \"%s\": %s", name, strerror (errno));
	mm_log(tmp,ERROR);
	return NIL;
    }

    /* There are two ways in which the name can come here, either as a 
       full path or not. If it is not a full path it can come in two ways, 
       either as a file system path (Maildir/.Drafts) or as a maildir path 
       (INBOX.Drafts)
     */

     if(*name == '/')	/* full path */
	strcpy(tmp, name); /* do nothing */
     else{
	sprintf (tmp,"%s/%s%s%s", myrootdir (name),
	    strncmp (ucase (strcpy (tmp, name)), "INBOX", 5) 
		? name : maildirpath,
	    strncmp (ucase (strcpy (tmp, name)), "INBOX", 5) 
		? "" : (courier ? "/" : ""),
	    strncmp (ucase (strcpy (tmp, name)), "INBOX", 5) 
		? "" : (*(name+5) == MDSEPARATOR(courier) ? name+5 : ""));
     }

    return tmp[0] ? T : NIL;
}

/* This function is given a full path for a mailbox and returns
 * if it is a valid maildir transformed to canonical notation
 */
int
is_valid_maildir (char **name)
{
  if (!strncmp(*name, myrootdir (*name), strlen(myrootdir(*name)))){
     (*name) += strlen(myrootdir(*name));
     if (**name == '/') (*name)++;
  }
  return maildir_valid(*name) ? T :  NIL;
}

/* Check validity of mailbox. This routine does not send errors to log, other
 *  routines calling this one may do so, though
 */ 

DRIVER *maildir_valid (char *name)
{
   char tmpname[MAILTMPLEN];

   maildir_file_path(name, tmpname);
   
   return maildir_valid_dir(tmpname) 
		? (IS_COURIER(name) ? &courierdriver : &maildirdriver) : NIL;
}
/*maildir fast */
void maildir_fast (MAILSTREAM *stream,char *sequence,long flags)
{
  unsigned long i;
  MESSAGECACHE *elt;
                                /* get sequence */
  if (stream && LOCAL && ((flags & FT_UID) ?
                          mail_uid_sequence (stream,sequence) :
                          mail_sequence (stream,sequence)))
    for (i = 1; i <= stream->nmsgs; i++) {
      if ((elt = mail_elt (stream,i))->sequence && (elt->valid = T) &&
          !(elt->day && elt->rfc822_size)) {
        ENVELOPE **env = NIL;
        ENVELOPE *e = NIL;
        if (!stream->scache) env = &elt->private.msg.env;
        else if (stream->msgno == i) env = &stream->env;
        else env = &e;
        if (!*env || !elt->rfc822_size) {
          STRING bs;
          unsigned long hs;
          char *ht = (*stream->dtb->header) (stream,i,&hs,NIL);

          if (!*env) rfc822_parse_msg (env,NIL,ht,hs,NIL,BADHOST,
                                       stream->dtb->flags);
          if (!elt->rfc822_size) {
            (*stream->dtb->text) (stream,i,&bs,FT_PEEK);
            elt->rfc822_size = hs + SIZE (&bs) - GETPOS (&bs);
          }
        }

        if (!elt->day && *env && (*env)->date)
          mail_parse_date (elt,(*env)->date);

        if (!elt->day) elt->day = elt->month = 1;
        mail_free_envelope (&e);
      }
    }
}


/* 
 * return all files in a given directory. This is a separate call
 * so that if there are warnings during compilation this only appears once.
 */
unsigned long
maildir_scandir (char *name, struct direct ***flist, 
			unsigned long *nfiles, int *scand, int flag)
{
  struct stat sbuf;

  if (scand)
     *scand = -1;	/* assume error for safety */
  stat(name,&sbuf);	/* stat the containing directory */
  if (scand)
     *scand = scandir(name, flist, 
		(flag == CCLIENT ? maildir_select : courier_dir_select), 
		(flag == CCLIENT ? maildir_namesort : courier_dir_sort));
 *nfiles = (scand && (*scand > 0)) ? (unsigned long) *scand : 0L;

  return sbuf.st_ctime;
}

/* Does a message with given name exists (or was it removed)?
 * Returns: 1 - yes, such message exist,
 *	    0 - No, that message does not exist anymore
 *
 * Parameters: stream, name of mailbox, new name if his message does not
 *		exist.
 */

int maildir_message_exists(MAILSTREAM *stream, char *name, char *newfile)
{
  char tmp[MAILTMPLEN];
  int gotit = NIL;
  DIR *dir;
  struct direct *d;
  struct stat sbuf;

  /* First check directly if it exists, if not there, look for it */
  sprintf(tmp,"%s/%s", LOCAL->curdir, name);
  if ((stat(tmp, &sbuf) == 0) && ((sbuf.st_mode & S_IFMT) == S_IFREG))
    return T;

  if (!(dir = opendir (LOCAL->curdir)))
     return NIL;

  while ((d = readdir(dir)) && gotit == NIL){
    if (d->d_name[0] == '.')
      continue;
    if (same_maildir_file(d->d_name, name)){
	  gotit = T;
	  strcpy(newfile, d->d_name);
    }
  }
  closedir(dir);
  return gotit;
}

/* Maildir open */
 
MAILSTREAM *maildir_open (MAILSTREAM *stream)
{
  char tmp[MAILTMPLEN];
  struct stat sbuf;

  if (!stream) return &maildirproto;
  if (stream->local) fatal ("maildir recycle stream");
  md_domain_name();    /* get domain name for maildir files in mdlocaldomain */
  stream->uid_last  = stream->uid_validity = 0;
  if (!stream->rdonly){
     stream->perm_seen = stream->perm_deleted = stream->perm_flagged = 
	stream->perm_answered = stream->perm_draft = T;
  }
  stream->uid_validity = time(0);
  stream->local = (MAILDIRLOCAL *)fs_get (sizeof (MAILDIRLOCAL));
  memset(LOCAL, 0, sizeof(MAILDIRLOCAL));
  LOCAL->fd = -1;

  LOCAL->courier = IS_COURIER(stream->mailbox);
  strcpy(tmp, stream->mailbox);
  if (maildir_file_path (stream->mailbox, tmp))
     LOCAL->dir = cpystr (tmp);
  if (LOCAL->dir){
     MDFLD(tmp, LOCAL->dir, Cur);
     LOCAL->curdir = cpystr (tmp);
     if (stat (LOCAL->curdir,&sbuf) < 0) {
         sprintf (tmp,"Can't open folder %s: %s",
				stream->mailbox,strerror (errno));
         mm_log (tmp,ERROR);
	 maildir_close(stream, 0);
        return NIL;
     }
  }

  if(maildir_file_path (stream->mailbox, tmp)){
    fs_give ((void **) &stream->mailbox);
    stream->mailbox = cpystr(tmp);
  }

  LOCAL->buf = (char *) fs_get ((LOCAL->buflen = MAXMESSAGESIZE) + 1);
  stream->sequence++;
  stream->nmsgs = stream->recent = 0;

  maildir_parse_folder(stream, 1);

  return stream;
}

/* Maildir initial parsing of the folder */
void
maildir_parse_folder (MAILSTREAM *stream, int full)
{
   unsigned long total;

   if (!stream)		/* what??? */
      return;

   MM_CRITICAL(stream);

   /* Scan old messages first, escoba! */
   total = LOCAL ? maildir_parse_dir(stream, 0L, Cur, full) 
		 : stream->nmsgs;
   stream->nmsgs = LOCAL ? maildir_parse_dir(stream, total, New, full) 
			 : stream->nmsgs;

   MM_NOCRITICAL(stream);
}

/* Return the number of messages in the directory, while filling the
 * elt structure.
 */

unsigned long
maildir_parse_dir(MAILSTREAM *stream, unsigned long nmsgs,
		  DirNamesType dirtype, int full)
{
   char tmp[MAILTMPLEN], tmp2[MAILTMPLEN], file[MAILTMPLEN], 
	newfile[MAILTMPLEN], *mdstr;
   struct direct **names = NIL;
   struct stat sbuf;
   unsigned long i, j = 0L, nfiles, last_scan;
   unsigned long recent = stream ? stream->recent : 0L;
   int d = 0, f = 0, r = 0, s = 0, t = 0;
   int k, we_compute, in_list, scan_err;
   int silent = stream ? stream->silent : NIL;
   MESSAGECACHE *elt;

   MDFLD(tmp, LOCAL->dir, dirtype);
   if (access (tmp, R_OK|W_OK|X_OK) != 0){
      maildir_abort(stream);
      return stream->nmsgs;
   }

   MDFLD(tmp, LOCAL->dir, Cur);
   if (dirtype != New && 
	(stat(tmp, &sbuf) < 0 || sbuf.st_ctime == LOCAL->scantime))
      return stream->nmsgs;

   MDFLD(tmp, LOCAL->dir, dirtype);
   last_scan =  maildir_scandir (tmp, &names, &nfiles, &scan_err, CCLIENT);
   if (dirtype == Cur)
      LOCAL->scantime = last_scan;

   if (scan_err < 0){
	maildir_abort(stream);
	return nmsgs;
   }

   if (dirtype == Cur)
      for (i = 1L; i <= stream->nmsgs;){
	elt = mail_elt(stream,  i);
	in_list = elt && elt->maildirp && nfiles > 0L
		  ? (MDPOS(elt) < nfiles 
		    ? same_maildir_file(MDFILE(elt), names[MDPOS(elt)]->d_name)
		    : NIL)
		    || maildir_message_in_list(MDFILE(elt), names, 0L, 
						nfiles - 1L, &MDPOS(elt))
		  : NIL;
	if (!in_list){
	   if (elt->maildirp)
	      maildir_free_file ((void **) &elt->maildirp);

	   if (elt->recent) --recent;
	   mail_expunged(stream,i);
	}
	else i++;
      }

   stream->silent = T;
   for (we_compute = 0, i = 1L; i <= nfiles; i++){
      unsigned long pos, n;
      mail_exists(stream, i + nmsgs);
      elt = mail_elt(stream, i + nmsgs);
      if (elt && elt->maildirp)
         pos = MDPOS(elt);	/* use data we found above */
      else{
        if (full)
           pos = i - 1;		/* first time, use sequence number */
        else{
           for (n = 0L ; (n < nfiles) && !names[n] ; n++);
           pos = n;		/* nfiles > stream->nmsgs!!, assign one */
        }
      }
      if (dirtype == New) elt->recent = T;
      if (!elt->private.uid){
	 elt->private.uid = stream->uid_last + 1;
	 stream->uid_validity = time(0);
      }
      if (stream->uid_last < elt->private.uid) 
	  stream->uid_last = elt->private.uid;

      maildir_getflag(names[pos]->d_name, &d, &f, &r ,&s, &t);
      if (elt->maildirp)
	 maildir_free_file_only ((void **)&elt->maildirp);
      else{
	 maildir_get_file((MAILDIRFILE **)&elt->maildirp);
	 we_compute++;
      }
      MDFILE(elt) = cpystr(names[pos]->d_name);
      MDPOS(elt)  = pos;

      if (elt->draft != d || elt->flagged != f || 
	elt->answered != r || elt->seen != s || elt->deleted != t){
	   elt->draft = d; elt->flagged = f; elt->answered = r;
	   elt->seen  = s; elt->deleted = t;
	   if (!we_compute && !stream->rdonly)
	      MM_FLAGS(stream, i+nmsgs);
      }
      maildir_get_date(stream, i+nmsgs);
      elt->valid = T;
      if (dirtype == New && !stream->rdonly){ /* move new messages to cur */
	 sprintf (file,"%s/%s", tmp, names[pos]->d_name);
	 if (stat (file,&sbuf) == 0 && S_ISREG (sbuf.st_mode)){
	    strcpy(tmp2,names[pos]->d_name);
	    if ((mdstr = strstr (names[pos]->d_name,MDSEP(3)))
		|| (mdstr = strstr (names[pos]->d_name,MDSEP(2)))){ /* Grrr */
	       *(mdstr+1) = '2';
	       sprintf (newfile,"%s/%s",LOCAL->curdir,names[pos]->d_name);
	    }
	    else{
	       sprintf (newfile,"%s/%s%s",LOCAL->curdir,names[pos]->d_name,MDSEP(2));
	       strcat(tmp2, MDSEP(2));
	    }
	    if (link (file,newfile) < 0){
	       mm_log("Unable to read new mail!",WARN);
	    }
	    else{
	      unlink (file);
	      j++;	/* success!, count it! */
	    }
	    maildir_free_file_only((void **)&elt->maildirp);
	    MDFILE(elt) = cpystr(tmp2);
	    MDSIZE(elt) = sbuf.st_size;
	    MDMTIME(elt) = sbuf.st_mtime;
	    maildir_get_date(stream, i + nmsgs);
	 }
      }
      fs_give((void **)&names[pos]);
   }
   if(names)
      fs_give((void **) &names);
   stream->silent = silent;
   if (dirtype == New && stream->rdonly)
      j = nfiles;
   mail_exists(stream, nmsgs  + ((dirtype == New) ? j : nfiles));
   mail_recent(stream, recent + ((dirtype == New) ? j : 0));

   return (nmsgs  + (dirtype == New ? j : nfiles));
}

long maildir_ping (MAILSTREAM *stream)
{
  maildir_parse_folder(stream, 0);
  return stream && LOCAL ? T : NIL;
}

int maildir_select (struct direct *name)
{
 return (name->d_name[0] != '.');
}

/*
 * Unfortunately, there is no way to sort by arrival in this driver, this
 * means that opening a folder in this driver using the scandir function
 * will always make this driver slower than any driver that has a natural
 * way of sorting by arrival (like a flat file format, "mbox", "mbx", etc).
 */

int maildir_namesort (const void *d1,const  void *d2)
{
  const struct direct **e1, **e2;

  e1 = (const struct direct **)d1;
  e2 = (const struct direct **)d2;

  return comp_maildir_file((char*)(*e1)->d_name, (char *)(*e2)->d_name);
}

/* Maildir close */

void maildir_close (MAILSTREAM *stream, long options)
{
  MESSAGECACHE *elt;
  unsigned long i;
  int silent = stream ? stream->silent : 0;
  mailcache_t mc = (mailcache_t) mail_parameters (NIL,GET_CACHE,NIL);

  if (!stream) return;

  for (i = 1; i <= stream->nmsgs; i++)
    if ((elt = (MESSAGECACHE *) (*mc) (stream,i,CH_ELT)) && elt->maildirp)
      maildir_free_file ((void **) &(elt->maildirp));
  stream->silent = T;
  if (options & CL_EXPUNGE) maildir_expunge (stream);
  maildir_abort(stream);
  if (mdfpath) fs_give((void **)&mdfpath);
  stream->silent = silent;
}

void maildir_check (MAILSTREAM *stream)
{
  if (maildir_ping (stream)) mm_log ("Check completed",(long) NIL);   
}

long maildir_text (MAILSTREAM *stream,unsigned long msgno,STRING *bs, long flags)
{
  char tmp[MAILTMPLEN];
  unsigned long i;
  MESSAGECACHE *elt;
  char *s;
                                /* UID call "impossible" */
  if (flags & FT_UID || !LOCAL) return NIL;
  elt = mail_elt (stream,msgno);
  sprintf (tmp,"%s/%s",LOCAL->curdir, MDFILE(elt));
  if (LOCAL->fd < 0)	/* if file closed ? */
     LOCAL->fd = open(tmp,O_RDONLY,NIL);

  if (LOCAL->fd < 0 && (errno == EACCES || errno == ENOENT)){
     INIT (bs, mail_string, "", 0);
     elt->rfc822_size = 0L;
     return NIL;
  }

  if (LOCAL->dirty == 0)
     MM_FLAGS(stream, elt->msgno);

  s = maildir_text_work(stream, elt, &i, flags);
  INIT (bs, mail_string, s, i);
  return T;
}

char *maildir_text_work (MAILSTREAM *stream,MESSAGECACHE *elt,
                      unsigned long *length,long flags)
{
  FDDATA d;
  STRING bs;
  char *s,*t,*tl,tmp[CHUNK];
  unsigned long msgno = elt->msgno;
  static int try = 0;

  if (length)
     *length = 0L;
  LOCAL->buf[0] = '\0';

  sprintf (tmp,"%s/%s",LOCAL->curdir, MDFILE(elt));
  if (LOCAL->fd < 0)	/* if file closed ? */
     LOCAL->fd = open(tmp,O_RDONLY,NIL);

  if (LOCAL->fd < 0){		/* flag change? */
      if (try < 5){
	try++;
	if (maildir_update_elt_maildirp(stream, msgno) > 0)
	  try = 0;
	return maildir_text_work(stream, mail_elt(stream, msgno),length, flags);
      }
      try = 0;
      return NULL;
  }

  lseek (LOCAL->fd, elt->private.msg.text.offset,L_SET);

  if (flags & FT_INTERNAL) {    /* initial data OK? */
    if (elt->private.msg.text.text.size > LOCAL->buflen) {
      fs_give ((void **) &LOCAL->buf);
      LOCAL->buf = (char *) fs_get ((LOCAL->buflen =
                                     elt->private.msg.text.text.size) + 1);
    }
    read (LOCAL->fd,LOCAL->buf,elt->private.msg.text.text.size);
    LOCAL->buf[*length = elt->private.msg.text.text.size] = '\0';
  }
  else {
    if (elt->rfc822_size > LOCAL->buflen) {
      fs_give ((void **) &LOCAL->buf);
      LOCAL->buf = (char *) fs_get ((LOCAL->buflen = elt->rfc822_size) + 1);
    }
    d.fd = LOCAL->fd;           /* yes, set up file descriptor */
    d.pos = elt->private.msg.text.offset;
    d.chunk = tmp;              /* initial buffer chunk */
    d.chunksize = CHUNK;
    INIT (&bs,fd_string,&d,elt->private.msg.text.text.size);
    for (s = LOCAL->buf; SIZE (&bs);) switch (CHR (&bs)) {
    case '\r':                  /* carriage return seen */
      *s++ = SNX (&bs);         /* copy it and any succeeding LF */
      if (SIZE (&bs) && (CHR (&bs) == '\n')) *s++ = SNX (&bs);
      break;
    case '\n':
      *s++ = '\r';              /* insert a CR */
    default:
      *s++ = SNX (&bs);         /* copy characters */
    }
    *s = '\0';                  /* tie off buffer */
    *length = s - (char *) LOCAL->buf;   /* calculate length */
  }
  close(LOCAL->fd); LOCAL->fd = -1;
  return LOCAL->buf;
}

/* maildir parse, fill the elt structure... well not all of it... */
unsigned long maildir_parse_message(MAILSTREAM *stream, unsigned long msgno,
				    DirNamesType dirtype)
{
  char *b, *s, c;
  char tmp[MAILTMPLEN];
  struct stat sbuf;
  unsigned long i, len;
  int offset = 0, d, f, r, se, dt;
  MESSAGECACHE *elt;

  elt = mail_elt (stream,msgno);
  MSGPATH(tmp, LOCAL->dir, MDFILE(elt), dirtype);
  if(stat(tmp, &sbuf) == 0)
     MDSIZE(elt) = sbuf.st_size;

  maildir_get_date(stream, msgno);
  maildir_getflag(MDFILE(elt), &d, &f, &r ,&se, &dt);
  elt->draft = d; elt->flagged = f; elt->answered = r; elt->seen = se;
  elt->deleted = dt; elt->valid  = T;
  if (LOCAL->fd < 0)	/* if file closed ? */
     LOCAL->fd = open(tmp,O_RDONLY,NIL);

  if (LOCAL->fd >= 0){
	s = (char *) fs_get (MDSIZE(elt) + 1);
	read (LOCAL->fd,s,MDSIZE(elt));
	s[MDSIZE(elt)] = '\0';
	for (i = 0, b = s; *b && !(i && (*b == '\n')); i = (*b++ == '\n'));
	len = (*b ? ++b : b) - s;
	elt->private.msg.header.text.size = 
		elt->private.msg.text.offset = len;
	elt->private.msg.text.text.size = MDSIZE(elt) - len;
	for (i = 0, b = s, c = *b; b && c &&
	    ((c < '\016') && (((c == '\012') && ++i) ||
		((c == '\015') && (*++b == '\012') && (i +=2)))
	    || c); i++, c= *++b);
	elt->rfc822_size = i;
	fs_give ((void **) &s);
	close(LOCAL->fd); LOCAL->fd = -1;
  }
  return elt->rfc822_size;
}

int
maildir_update_elt_maildirp(MAILSTREAM *stream, unsigned long msgno)
{
     char tmp[MAILTMPLEN];
     struct direct **names = NIL;
     unsigned long i, nfiles, pos;
     int d = 0, f = 0 , r = 0, s = 0, t = 0, in_list, scan_err;
     MESSAGECACHE *elt;

     MDFLD(tmp, LOCAL->dir, Cur);

     maildir_scandir (tmp, &names, &nfiles, &scan_err, CCLIENT);

     elt = mail_elt (stream,msgno);

     in_list = nfiles > 0L
		? maildir_message_in_list(MDFILE(elt), names, 0L, 
					nfiles - 1L, &pos)
		: NIL;

     if (in_list && pos >= 0L && pos < nfiles
	 && !strcmp(MDFILE(elt), names[pos]->d_name)){
	in_list = NIL;
	maildir_abort(stream);
     }

     if (in_list && pos >= 0L && pos < nfiles){
	maildir_free_file_only((void **)&elt->maildirp);
	MDFILE(elt) = cpystr(names[pos]->d_name);
	maildir_getflag(MDFILE(elt), &d, &f, &r ,&s, &t);
	if (elt->draft != d || elt->flagged != f || 
	   elt->answered != r || elt->seen != s || elt->deleted != t){
	   elt->draft = d; elt->flagged = f; elt->answered = r;
	   elt->seen  = s; elt->deleted = t;
	   MM_FLAGS(stream, msgno);
        }
     }
     for (i = 0L; i < nfiles; i++)
	fs_give((void **) &names[i]);
     if (names)
	fs_give((void **) &names);
     return in_list ? 1 : -1;
}

/* Maildir fetch message header */

char *maildir_header (MAILSTREAM *stream,unsigned long msgno,
		unsigned long *length, long flags)
{
  char tmp[MAILTMPLEN], *s;
  MESSAGECACHE *elt;
  static int try = 0;

  if (length)
     *length = 0;
  if (flags & FT_UID || !LOCAL) return "";	/* UID call "impossible" */
  elt = mail_elt (stream,msgno);
  if(elt->private.msg.header.text.size == 0)
     maildir_parse_message(stream, msgno, Cur);

  sprintf (tmp,"%s/%s",LOCAL->curdir, MDFILE(elt));
  if (LOCAL->fd < 0)
     LOCAL->fd = open (tmp,O_RDONLY,NIL);

  if (LOCAL->fd < 0 && errno == EACCES){
     mm_log ("Message exists but can not be read. Envelope and body lost!",ERROR);
     return NULL;
  }

  if (LOCAL->fd < 0){			/* flag change? */
      if (try < 5){
	try++;
	if (maildir_update_elt_maildirp(stream, msgno) > 0)
	  try = 0;
	return maildir_header(stream, msgno, length, flags);
      }
      try = 0;
      return NULL;
  }

  if ((flags & FT_INTERNAL) &&
        (elt->private.msg.header.text.size > LOCAL->buflen)){
         fs_give ((void **) &LOCAL->buf);
         LOCAL->buf = (char *) fs_get ((LOCAL->buflen =
                                 elt->private.msg.header.text.size) + 1);
  }
  else
      s = (char *) fs_get(elt->private.msg.header.text.size+1);
  if (LOCAL->fd >= 0){
     read (LOCAL->fd, ((flags & FT_INTERNAL) ? (void *)LOCAL->buf : (void *)s),
                                      elt->private.msg.header.text.size);
     if (flags & FT_INTERNAL)
        LOCAL->buf[*length = elt->private.msg.header.text.size] = '\0';
     else{
        s[*length = elt->private.msg.header.text.size] = '\0';
        *length = strcrlfcpy (&LOCAL->buf,&LOCAL->buflen,s,
                          elt->private.msg.header.text.size);
        fs_give ((void **) &s);
     }
  }
  elt->private.msg.text.offset = elt->private.msg.header.text.size;
  elt->private.msg.text.text.size = MDSIZE(elt) - elt->private.msg.text.offset;
  if(s)
    fs_give((void **)&s);
  close(LOCAL->fd); LOCAL->fd = -1;
  return LOCAL->buf;
}

/* Maildir find list of subscribed mailboxes
 * Accepts: mail stream
 *	    pattern to search
 */

void maildir_list (MAILSTREAM *stream,char *ref, char *pat)
{
  char *s,test[MAILTMPLEN],file[MAILTMPLEN];
  long i = 0;

  if((!pat || !*pat) && maildir_canonicalize (test,ref,"*")
	&& maildir_valid_name(test)){	/* there is a #md/ leading here */
    for (i = 3; test[i] && test[i] != '/'; i++);
    if (s = strchr (test+i+1,'/')) *++s = '\0';
    else test[0] = '\0';
    mm_list (stream,'/',test, LATT_NOSELECT);
  }
  else if (maildir_canonicalize (test,ref,pat)) {
    if (test[3] == '/') {       /* looking down levels? */
                                /* yes, found any wildcards? */
      if (s = strpbrk (test,"%*")) {
                                /* yes, copy name up to that point */
        strncpy (file,test+4,i = s - (test+4));
        file[i] = '\0';         /* tie off */
      }
      else strcpy (file,test+4);/* use just that name then */
                                /* find directory name */
      if (s = strrchr (file, '/')){
        *s = '\0';              /* found, tie off at that point */
        s = file;
      }
                                /* do the work */
      if(IS_COURIER(test))
	courier_list_work (stream,s,test,0);
      else
	maildir_list_work (stream,s,test,0);
    }
                                /* always an INBOX */
    if (!compare_cstring (test,"#MD/INBOX"))
      mm_list (stream,NIL,"#MD/INBOX",LATT_NOINFERIORS);
    if (!compare_cstring (test,"#MC/INBOX"))
      mm_list (stream,NIL,"#MC/INBOX",LATT_NOINFERIORS);
  }
}

void courier_list (MAILSTREAM *stream,char *ref, char *pat)
{
/* I am too lazy to do anything. Do you care to ask maildir list, please?
   The real reason why this is a dummy function is because we do not want to
   see the same folder listed twice. 
*/
}

/* For those that want to hide things, we give them a chance to do so */
void *maildir_parameters (long function, void *value)
{
  void *ret = NIL;
  switch ((int) function) {
  case SET_INBOXPATH:
    if (myMdInboxDir) fs_give ((void **) &myMdInboxDir);
    myMdInboxDir = cpystr ((char *) value);
  case GET_INBOXPATH:
    if (!myMdInboxDir) myMdInboxDir = cpystr("Maildir");
    ret = (void *) myMdInboxDir;
    break;
  case SET_COURIERSTYLE:
    CourierStyle = (long) value;
  case GET_COURIERSTYLE:
    ret = (void *) CourierStyle;
    break;
  default:
    break;
  }
  return ret;
}

int maildir_create_folder(char *mailbox)
{
  char tmp[MAILTMPLEN], err[MAILTMPLEN];
  int i;

  for (i = Cur; i != EndDir; i++){
	MDFLD(tmp, mailbox, i);
	if (mkdir(tmp, 0700) && errno != EEXIST){ /* try to make new dir */
	    sprintf (err, "Can't create %s: %s", tmp, strerror(errno));
	    mm_log (err,ERROR);
	    return NIL;
	}
  }
  return T;
}

int maildir_create_work(char *mailbox, int loop)
{
  char *s, c, err[MAILTMPLEN], tmp[MAILTMPLEN], tmp2[MAILTMPLEN], mbx[MAILTMPLEN];
  int fnlen, i, create_dir = 0, courier, mv;
  struct stat sbuf;
  long style = (long) maildir_parameters(GET_COURIERSTYLE, NIL);

  courier = IS_COURIER(mailbox);
  strcpy(mbx, mailbox);
  mv = maildir_valid(mbx) ? 1 : 0;
  maildir_file_path(mailbox, tmp);
  if (mailbox[strlen(mailbox) - 1] == MDSEPARATOR(courier)){
      create_dir++;
      mailbox[strlen(mailbox) - 1] = '\0';
  }

  if(!loop && courier){
    if(mv){
       if(create_dir){
	  if(style == CCLIENT)
	   strcpy (err,"Can not create directory: folder exists. Create subfolder");
	  else
	   strcpy(err,"Folder and Directory already exist");
       }
       else
          strcpy (err, "Can't create mailbox: mailbox already exists");
    }
    else{
	if(create_dir)
	   strcpy(err, "Can not create directory. Cread folder instead");
	else
	  err[0] = '\0';
    }
    if(err[0]){
       mm_log (err,ERROR);
       return NIL;
    }
  }

  fnlen = strlen(tmp);
  if (s = strrchr(mailbox,MDSEPARATOR(courier))){
     c = *++s;
    *s = '\0';
    if ((stat(tmp,&sbuf) || ((sbuf.st_mode & S_IFMT) != S_IFDIR)) &&
        !maildir_create_work (mailbox, ++loop))
      return NIL;
    *s = c;
  }
  tmp[fnlen] = '\0';

  if (mkdir(tmp,0700) && errno != EEXIST)
     return NIL;

  if (create_dir)
     mailbox[fnlen] = '/';

  if (create_dir){
     if(style == CCLIENT){
	if(!courier){
	   FILE *fp = NULL;
	   sprintf(tmp2,"%s%s", tmp, MDDIR);
	   if ((fp = fopen(tmp2,"w")) == NULL){
	      sprintf (err,"Problem creating %s: %s", tmp2, strerror(errno));
              mm_log (err,ERROR);
              return NIL;
	   }
	   fclose(fp);
	}
     }
     return T;
  }
  else
     return maildir_create_folder(tmp);
}

long maildir_create (MAILSTREAM *stream,char *mailbox)
{
  char tmp[MAILTMPLEN], err[MAILTMPLEN];
  long rv;
  int create_dir;

  create_dir = mailbox ? 
		(mailbox[strlen(mailbox) - 1] == 
					MDSEPARATOR(IS_COURIER(mailbox))) : 0;
  maildir_file_path(mailbox, tmp);
  strcpy(tmp, mailbox);
  rv = maildir_create_work(mailbox, 0);
  strcpy(mailbox, tmp);
  if (rv == 0){
     sprintf (err,"Can't create %s %s",
		   (create_dir ? "directory" : "mailbox"), mailbox);
     mm_log (err,ERROR);
  }
  return rv ? 1L : 0L;
}

#define MAXTRY 10000
void maildir_flagmsg (MAILSTREAM *stream,MESSAGECACHE *elt)
{
  char oldfile[MAILTMPLEN],newfile[MAILTMPLEN],fn[MAILTMPLEN];
  char tmp[MAILTMPLEN];
  char *s;
  int ren, try = 0;

  LOCAL->dirty = 0;
  if (elt->valid){
     for (try = 1; try > 0 && try < MAXTRY; try++){
                                /* build the new filename */
	sprintf (oldfile,"%s/%s",LOCAL->curdir, MDFILE(elt));
	fn[0] = '\0';
	if ((ren = maildir_message_exists(stream, MDFILE(elt), fn)) == 0){
	    errno = ENOENT;
	    try = MAXTRY;
	}
	if (*fn)	/* new oldfile! */
	   sprintf (oldfile,"%s/%s",LOCAL->curdir,fn);
        if ((s = strchr (MDFILE(elt), FLAGSEP))) *s = '\0';
	sprintf (fn,"%s%s%s%s%s%s%s", MDFILE(elt), MDSEP(2),
		MDFLAG(Draft, elt->draft), MDFLAG(Flagged, elt->flagged),
		MDFLAG(Replied, elt->answered), MDFLAG(Seen, elt->seen),
		MDFLAG(Trashed, elt->deleted));
	sprintf (newfile,"%s/%s",LOCAL->curdir,fn);
        if (ren != 0 && rename (oldfile,newfile) >= 0)
	    try = -1;
     }

     if (try > 0){
       sprintf(oldfile,"Unable to write flags to disk: %s",
		(errno == ENOENT) ? "message is gone!" : strerror (errno));
       mm_log(oldfile,ERROR);
       LOCAL->dirty = 1;
       return;
     }
     maildir_free_file_only ((void **) &elt->maildirp);
     MDFILE(elt) = cpystr (fn);
  }
}

void maildir_expunge (MAILSTREAM *stream)
{
  MESSAGECACHE *elt;
  unsigned long i;
  unsigned long n = 0;
  unsigned long nmsgs = stream->nmsgs;
  unsigned long recent = stream->recent;
  char tmp[MAILTMPLEN];
  
  mm_critical (stream);		/* go critical */
  for (i = 1; i <= stream->nmsgs;)
    if ((elt = mail_elt (stream,i))->deleted){
      sprintf (tmp,"%s/%s",LOCAL->curdir, MDFILE(elt));
      if (unlink (tmp)) {/* try to delete the message */
	sprintf (tmp,"Expunge of message %ld failed, aborted: %s",i,
		 strerror (errno));
	if (!stream->silent)
	   mm_log (tmp,WARN);
	break;
      }
				/* free the cached filename */
      if (elt->maildirp)
	 maildir_free_file ((void **) &elt->maildirp);

      if (elt->recent) --recent;/* if recent, note one less recent message */
      mail_expunged (stream,i);	/* notify upper levels */
      n++;			/* count up one more expunged message */
    }
    else i++;
  if (n) {			/* output the news if any expunged */
    sprintf (tmp,"Expunged %ld messages",n);
    if (!stream->silent)
       mm_log (tmp,(long) NIL);
  }
  else 
    if (!stream->silent)
	mm_log ("No messages deleted, so no update needed",(long) NIL);
  mm_nocritical (stream);	/* release critical */
				/* notify upper level of new mailbox size */
  mail_exists (stream,stream->nmsgs);
  mail_recent (stream,recent);
}

long maildir_copy (MAILSTREAM *stream,char *sequence,char *mailbox,long options)
{
  STRING st;
  MESSAGECACHE *elt;
  unsigned long len;
  int fd;
  long i, length;
  struct stat sbuf;
  char tmp[MAILTMPLEN], flags[MAILTMPLEN], path[MAILTMPLEN], *s, *b;
				/* copy the messages */
  if ((options & CP_UID) ? mail_uid_sequence (stream, sequence) : 
  	mail_sequence (stream,sequence)) 
  for (i = 1; i <= stream->nmsgs; i++)
    if ((elt = mail_elt (stream,i))->sequence){
      sprintf (path,"%s/%s",LOCAL->curdir, MDFILE(elt));
      if (((fd = open (path,O_RDONLY,NIL)) < 0)	 
	  ||((!elt->rfc822_size && 
		((stat(path, &sbuf) < 0) || !S_ISREG (sbuf.st_mode)))))
	return NIL;
	if(!elt->rfc822_size)
	  MDSIZE(elt) = sbuf.st_size;
        s = (char *) fs_get(MDSIZE(elt) + 1);
        read (fd,s,MDSIZE(elt));
        s[MDSIZE(elt)] = '\0';
        close (fd);
	len = strcrlfcpy (&LOCAL->buf,&LOCAL->buflen, s, MDSIZE(elt));
        INIT (&st,mail_string, LOCAL->buf, len);
	elt->rfc822_size = len;
	fs_give ((void **)&s);

      flags[0] = flags[1] = '\0';
      if (elt->seen) strcat (flags," \\Seen");
      if (elt->draft) strcat (flags," \\Draft");
      if (elt->deleted) strcat (flags," \\Deleted");
      if (elt->flagged) strcat (flags," \\Flagged");
      if (elt->answered) strcat (flags," \\Answered");
      flags[0] = '(';         /* open list */
      strcat (flags,")");     /* close list */
      mail_date (tmp,elt);	/* generate internal date */
      if (!mail_append_full (NIL,mailbox,flags,tmp,&st))
        return NIL;
      if (options & CP_MOVE) elt->deleted = T;
    }
  return T;			/* return success */
}

long maildir_append (MAILSTREAM *stream,char *mailbox,append_t af,void *data)
{
  int fd;
  STRING *message;
  char c,*s, *flags, *date;
  char tmp[MAILTMPLEN],file[MAILTMPLEN],path1[MAILTMPLEN],path2[MAILTMPLEN];
  MESSAGECACHE elt;
  long i;
  long size = 0;
  long ret = LONGT;
  unsigned long uf;
  long f;
  static unsigned int transact = 0;

  if (!maildir_valid(mailbox)) {
    sprintf (tmp,"Not a valid Maildir mailbox: %s",mailbox);
    mm_log (tmp,ERROR);
    return NIL;
  }

 if (!*mdlocaldomain)
     md_domain_name();    /* get domain name for maildir files in mdlocaldomain now! */

  if (!(*af) (stream,data,&flags,&date,&message)) return NIL;

  mm_critical (stream);		/* go critical */
  do {
    if (!SIZE (message)) {      /* guard against zero-length */
      mm_log ("Append of zero-length message",ERROR);
      ret = NIL;
      break;
    }
    if (date && !mail_parse_date(&elt,date)){
        sprintf (tmp,"Bad date in append: %.80s",date);
        mm_log (tmp,ERROR);
        ret = NIL;
        break;
    }
    f = mail_parse_flags (stream,flags,&uf);
				/* build file name we will use */
    sprintf (file,"%u.%d_%09u.%s%s%s%s%s%s",
	   time (0),getpid (),transact++,mdlocaldomain, (f ? MDSEP(2) : ""),
		MDFLAG(Draft, f&fDRAFT), MDFLAG(Flagged, f&fFLAGGED),
		MDFLAG(Replied, f&fANSWERED), MDFLAG(Seen, f&fSEEN));
				/* build tmp file name */
    if (maildir_file_path(mailbox, tmp))
       MSGPATH(path1, tmp, file, Tmp);

    if ((fd = open (path1,O_WRONLY|O_CREAT|O_EXCL,S_IREAD|S_IWRITE)) < 0) {
       sprintf (tmp,"Can't open append mailbox: %s",strerror (errno));
       mm_log (tmp,ERROR);
       return NIL;
    }
    for (size = 0,i = SIZE (message),s = (char *) fs_get (i + 1); i; --i)
      if ((c = SNX (message)) != '\015') s[size++] = c;
    if ((write (fd,s,size) < 0) || fsync (fd)) {
	unlink (path1);		/* delete message */
	sprintf (tmp,"Message append failed: %s",strerror (errno));
	mm_log (tmp,ERROR);
	ret = NIL;
    }
    fs_give ((void **) &s);	/* flush the buffer */
    close (fd);			/* close the file */
				/* build final filename to use */
    if (maildir_file_path(mailbox, tmp))
	MSGPATH(path2, tmp, file, New);
    if (link (path1,path2) < 0) {
       sprintf (tmp,"Message append failed: %s",strerror (errno));
       mm_log (tmp,ERROR);
       ret = NIL;
    }
    unlink (path1);

    if (ret)
     if (!(*af) (stream,data,&flags,&date,&message)) ret = NIL;

  } while (ret && message);	/* write the data */
  
  mm_nocritical (stream);	/* release critical */
  return ret;
}

long maildir_delete (MAILSTREAM *stream,char *mailbox)
{
  DIR *dirp;
  struct direct *d;
  int i, remove_dir = 0, mddir = 0, rv, error = 0;
  char tmp[MAILTMPLEN],tmp2[MAILTMPLEN], realname[MAILTMPLEN];
  struct stat sbuf;
  char *mdpath = mdirpath();
  int courier = IS_COURIER(mailbox);

  if (mailbox[strlen(mailbox) - 1] == MDSEPARATOR(courier)){
      remove_dir++;
      mailbox[strlen(mailbox) -1] = '\0';
  }

  if (!maildir_valid(mailbox)){
      maildir_file_path(mailbox, tmp);
      if (stat(tmp, &sbuf) < 0 || !S_ISDIR(sbuf.st_mode)){
        sprintf(tmp,"Can not remove %s", mailbox);
	error++;
      }
  }

  if (!error && remove_dir && !maildir_dir_is_empty(mailbox)){
     sprintf(tmp,"Can not remove directory %s/: directory not empty", mailbox);
     error++;
  }

  if(error){
     mm_log (tmp,ERROR);
     return NIL;
  }

  maildir_close(stream,0);	/* even if stream was NULL */

  maildir_file_path(mailbox, realname);

  if (remove_dir){
     sprintf(tmp,"%s/%s", realname, MDDIR);
     if ((rv = stat (tmp,&sbuf)) == 0 && S_ISREG(sbuf.st_mode))
	rv = unlink(tmp);
     else if (errno == ENOENT)
	rv = 0;
     if (rv != 0){
	sprintf(tmp,"Can not remove %s/%s: %s", tmp2, MDDIR, strerror(errno));
	mm_log (tmp,ERROR);
	return NIL;
     }
     if (!maildir_valid(realname) && rmdir(realname) != 0){
	sprintf(tmp,"Can not remove %s/: %s", mailbox, strerror(errno));
	mm_log (tmp,ERROR);
	return NIL;
     }
     return T;
  }
  /* else remove just the folder. Remove all hidden files, except MDDIR */
  for (i = Cur; i != EndDir; i++){
      MDFLD(tmp, realname, i);

      if (!(dirp = opendir (tmp))){
	  sprintf(tmp,"Can not read %s/: %s", mailbox, strerror(errno));
	  mm_log (tmp,ERROR);
	  return NIL;
      }

      while (d = readdir(dirp)){
	 if (strcmp(d->d_name, ".") && strcmp(d->d_name,"..")){
	    sprintf(tmp2,"%s/%s", tmp, d->d_name);
	    if (unlink(tmp2) != 0){
	       sprintf(tmp2,"Can not remove %s: %s", mailbox, strerror(errno));
	       mm_log (tmp2,ERROR);
	       return NIL;
	    }
	 }
      }
      closedir(dirp);
      if (rmdir(tmp) != 0){
	 sprintf(tmp,"Can not remove %s: %s", mailbox, strerror(errno));
	 mm_log (tmp,ERROR);
	 return NIL;
      }
  }
  /* 
   * ok we have removed all subdirectories of the folder mailbox, Remove the
   * hidden files.
   */

  if (!(dirp = opendir (realname))){
	sprintf(tmp,"Can not read %s/: %s", realname, strerror(errno));
	mm_log (tmp,ERROR);
	return NIL;
  }

  while (d = readdir(dirp)){
	if (strcmp(d->d_name, ".") && strcmp(d->d_name,"..")
		&& !strcmp(d->d_name, MDDIR)){
	   sprintf(tmp,"%s/%s", realname, d->d_name);
	   mddir++;
	   if (unlink(tmp) != 0)
	      error++;
	}
  }
  closedir(dirp);
  if (error || 
	 (maildir_dir_is_empty(mailbox) && mddir == 0 && rmdir(realname) < 0)){
        sprintf(tmp,"Can not remove folder %s: %s", mailbox, strerror(errno));
        mm_log (tmp,ERROR);
        return NIL;
  }
  return T;
}

long maildir_rename (MAILSTREAM *stream, char *old, char *new)
{
  struct direct **names = NIL;
  struct stat sbuf;
  char pathname[MAILTMPLEN], dir[MAILTMPLEN], curdir[MAILTMPLEN];
  char tmp[MAILTMPLEN],tmpnew[MAILTMPLEN], realold[MAILTMPLEN];
  char realnew[MAILTMPLEN],realpat[MAILTMPLEN], realname[MAILTMPLEN];
  char *pat, *endold, c;
  char *maildirpath = mdirpath();
  int courier = IS_COURIER(old) && IS_COURIER(new);
  int scand, i, j, rv = T;
  unsigned long ndir;
  COURIER_S *cdir;

  if((IS_COURIER(old) || IS_COURIER(new)) && !courier){
    sprintf (tmp,"Can't rename mailbox %s to %s",old, new);
    mm_log (tmp,ERROR);
    return NIL;
  }

  if (!maildir_valid(old)){
    sprintf (tmp,"Can't rename mailbox %s: folder not in maildir format",old);
    mm_log (tmp,ERROR);
    return NIL;
  }
  maildir_file_path(old, realold);
  if (!maildir_valid_name(new) && new[0] == '#'){
    sprintf (tmp,"Can't rename mailbox %s: folder not in maildir format",new);
    mm_log (tmp,ERROR);
    return NIL;
  }
  maildir_file_path(new, realnew);
  if (access(tmpnew,F_OK) == 0){ 	/* new mailbox name must not exist */
    sprintf (tmp,"Can't rename to mailbox %s: destination already exists",new);
    mm_log (tmp,ERROR);
    return NIL;
  }

  if(!courier){
    if (rename (realold,realnew)){	/* try to rename the directory */
       sprintf (tmp,"Can't rename mailbox %s to %s: %s",old, new,
							strerror(errno));
       mm_log (tmp,ERROR);
       return NIL;
    }
    return T;	/* return success */
  }

  cdir = courier_list_dir(old);
  for (i = 0; cdir && i < cdir->total; i++){
      if(strstr(cdir->data[i]->name, old)){
	sprintf(tmp,"%s%s", new, cdir->data[i]->name+strlen(old));
	maildir_file_path(cdir->data[i]->name, realold);
	maildir_file_path(tmp, realnew);
	if (rename (realold,realnew)){
	   sprintf (tmp,"Can't rename mailbox %s to %s: %s",old, new,
							strerror(errno));
	   mm_log (tmp,ERROR);
	   rv = NIL;
	}
    }
  }
  courier_free_cdir(&cdir);
  return rv;
}

long maildir_sub (MAILSTREAM *stream,char *mailbox)
{
  return sm_subscribe (mailbox);
}

long maildir_unsub (MAILSTREAM *stream,char *mailbox)
{
  return sm_unsubscribe (mailbox);
}

void maildir_lsub (MAILSTREAM *stream,char *ref,char *pat)
{
  void *sdb = NIL;
  char *s, 	test[MAILTMPLEN];
                                /* get canonical form of name */
  if (maildir_canonicalize (test,ref,pat) && (s = sm_read (&sdb))) {
    do if (pmatch_full (s,test,'/')) mm_lsub (stream,'/',s,NIL);
    while (s = sm_read (&sdb)); /* until no more subscriptions */
  }
}

long maildir_canonicalize (char *pattern,char *ref,char *pat)
{
  if (ref && *ref) {            /* have a reference */
    strcpy (pattern,ref);       /* copy reference to pattern */
                                /* # overrides mailbox field in reference */
    if (*pat == '#') strcpy (pattern,pat);
                                /* pattern starts, reference ends, with / */
    else if ((*pat == '/') && (pattern[strlen (pattern) - 1] == '/'))
      strcat (pattern,pat + 1); /* append, omitting one of the period */
                                                                                
    else strcat (pattern,pat);  /* anything else is just appended */
  }
  else strcpy (pattern,pat);    /* just have basic name */
  return (maildir_valid_name(pattern));
}

void maildir_list_work (MAILSTREAM *stream,char *dir,char *pat,long level)
{
  DIR *dp;
  struct direct *d;
  struct stat sbuf;
  char *cp,*np, curdir[MAILTMPLEN],name[MAILTMPLEN], tmp[MAILTMPLEN];
  char realpat[MAILTMPLEN];
  int i;
  char *maildirpath = mdirpath();

  sprintf(curdir,"%s/%s/", myrootdir(pat), dir ? dir : maildirpath);
  if (dp = opendir (curdir)){ 
     if (dir) sprintf (name,"%s%s/",MDPREFIX(CCLIENT),dir);
     else strcpy (name, pat);

     if (level == 0 && !strpbrk(pat,"%*")){
	if(maildir_valid(pat)){
	  i =  maildir_contains_folder(pat, NULL)
		? LATT_HASCHILDREN
		: (maildir_is_dir(pat, NULL)
			     ? LATT_HASNOCHILDREN : LATT_NOINFERIORS);
	  maildir_file_path(pat, realpat);
	  i +=  maildir_any_new_msgs(realpat) 
			? LATT_MARKED : LATT_UNMARKED;
	  mm_list (stream,'/', pat, i);
	}
	else
	   if(pat[strlen(pat) - 1] == '/')
	     mm_list (stream,'/', pat, LATT_NOSELECT);
     }

     while (d = readdir (dp))
	if(strcmp(d->d_name, ".") && strcmp(d->d_name,"..")
		&& strcmp(d->d_name, MDNAME(Cur)) 
		&& strcmp(d->d_name, MDNAME(Tmp)) 
		&& strcmp(d->d_name, MDNAME(New))){

	  if (dir) sprintf (tmp,"%s%s", name,d->d_name);
	  else strcpy(tmp, d->d_name);

	  if(pmatch_full (tmp, pat,'/')){
	     sprintf(tmp,"%s/%s/%s", myrootdir(d->d_name), 
				(dir ? dir : maildirpath), d->d_name);
	     if(stat (tmp,&sbuf) == 0 
		   && ((sbuf.st_mode & S_IFMT) == S_IFDIR)){
	       if (dir) sprintf (tmp,"%s%s", name,d->d_name);
	       else strcpy(tmp, d->d_name);
               i = maildir_valid(tmp)
			? (maildir_contains_folder(dir, d->d_name)
			  ? LATT_HASCHILDREN
			  : (maildir_is_dir(dir, d->d_name)
			     ? LATT_HASNOCHILDREN : LATT_NOINFERIORS))
			: LATT_NOSELECT;
	       i +=  maildir_any_new_msgs(tmp)
			    ? LATT_MARKED : LATT_UNMARKED;
	       mm_list (stream,'/',tmp, i);
	       strcat (tmp, "/");
	       if(dmatch (tmp, pat,'/') &&
                 (level < (long) mail_parameters (NIL,GET_LISTMAXLEVEL,NIL))){
		   sprintf(tmp,"%s/%s",dir,d->d_name);
 		   maildir_list_work (stream,tmp,pat,level+1);
	       }
	     }
	  }
       }
  }
  closedir (dp);
}

void courier_list_work (MAILSTREAM *stream, char *dir, char *pat, long level)
{
  unsigned long ndir;
  struct direct **names = NIL;
  struct stat sbuf;
  char c, d, curdir[MAILTMPLEN],name[MAILTMPLEN], tmp[MAILTMPLEN];
  char realname[MAILTMPLEN], realpat[MAILTMPLEN] = {'\0'}, pathname[MAILTMPLEN];
  char fakepat[MAILTMPLEN];
  int i, j, scand;
  long style = (long) maildir_parameters(GET_COURIERSTYLE, NIL);
  char *maildirpath = mdirpath();
  COURIER_S *cdir;

  if(!strpbrk(pat,"%*")){	/* a mailbox */
     maildir_file_path(pat, curdir);
     i = strlen(curdir) - 1;
     if(curdir[i] == '/')
       curdir[i] = '\0';
     cdir = courier_list_dir(curdir);
     if(cdir){
	if(maildir_valid(curdir)){
	  if(cdir->data[0]->name)
	    j = LATT_HASCHILDREN;
	  else
	    j = (style == COURIER) ? LATT_HASNOCHILDREN : LATT_NOINFERIORS;
	}
	else
	    j = LATT_NOSELECT;
        j += maildir_any_new_msgs(curdir) ? LATT_MARKED : LATT_UNMARKED;
        mm_list (stream, '.',pat, j);
        courier_free_cdir(&cdir);
     }
     return;
  }

  strcpy(tmp,pat + 4);	/* a directory */
  j = strlen(pat) - 1;
  maildir_file_path(pat, realpat);
  c = pat[j];
  pat[j] = '\0';
  realname[0] = '\0';
  if(dir)
    maildir_file_path(dir, realname);
  sprintf(curdir,"%s%s%s/%s", (dir ? "" : myrootdir(pat)), (dir ? "" : "/"),
		(dir ? realname : maildirpath),	(dir ? "" : "."));
  sprintf(tmp, "%s%s/.", MDPREFIX(COURIER), dir ? dir : maildirpath);
  if (level == 0 && tmp && pmatch_full (tmp, realpat, '.'))
     mm_list (stream,'.', tmp, LATT_NOSELECT);

  cdir = courier_list_dir(pat);
  pat[j] = c;
  for (i = 0; cdir && i < cdir->total; i++)
   if(pmatch_full (cdir->data[i]->name, pat, '.')){
      sprintf(tmp, "%s.", cdir->data[i]->name);
      courier_list_info(&cdir, tmp, i);
      mm_list (stream,'.',cdir->data[i]->name, cdir->data[i]->attribute);
   }
  courier_free_cdir(&cdir);
}

int 
same_maildir_file(char *name1, char *name2)
{
 char tmp1[MAILTMPLEN], tmp2[MAILTMPLEN];
 char *s;

 strcpy(tmp1, name1 ? name1 : "");
 strcpy(tmp2, name2 ? name2 : "");
 if (s = strchr(tmp1, FLAGSEP))
   *s = '\0';
 if (s = strchr(tmp1, SIZESEP))
   *s = '\0';
 if (s = strchr(tmp2, FLAGSEP))
   *s = '\0';
 if (s = strchr(tmp2, SIZESEP))
   *s = '\0';

 return !strcmp(tmp1, tmp2);
}

unsigned long antoul(char *seed)
{
  int i, error = 0;
  unsigned long val = 0L, rv = 0L;
  char c;
 if(!seed)
   return 0L;
 i = strlen(seed);
 for(i= strlen(seed)-1; i >= 0; i--){
    c = seed[i];
    if (c >= 'A' && c <= 'Z') val = c - 'A';
    else if (c >= 'a' && c <= 'z') val = c - 'a' + 26;
    else if (c >= '0' && c <= '9') val = c - '0' + 26 + 26; 
    else if (c == '-') val = c - '-' + 26 + 26 + 10;
    else if (c == '_') val = c - '_' + 26 + 26 + 10 + 1;
    else error++;
    rv = val + (rv << 6);
 }
  return error ? 0L : rv;
}

unsigned long mdfntoul (char *name, char **s)
{
  unsigned long t;
  char *r, *p, last;

  if(r = strchr(name,'.'))
     *r = '\0';
  t = strtoul(name, &p, 10);
  if(p && p - name < 8){
     last = *p;
     *p = '\0';
     t = antoul(name);
     *p = last;
  }
  *r = '.';
  if (s) *s = p ? p : r;
  return t;
}

int comp_maildir_file(char *name1, char *name2)
{
  unsigned long t1, t2;
  char *s1, *s2;
  int i,j;

  if (!(name1 && *name1))
     return (name2 && *name2) ? (*name2 == FLAGSEP ? 0 : -1) : 0;

  if (!(name2 && *name2))
     return (name1 && *name1) ? (*name1 == FLAGSEP ? 0 : 1) : 0;

   if(!strcmp(name1,name2))
      return 0;

  t1 = mdfntoul(name1, &s1);
  t2 = mdfntoul(name2, &s2);

  if (t1 < t2)
     return -1;

  if (t1 > t2)
    return 1;

  i = strchr(name1,'.') - name1 + 1;
  j = strchr(name2,'.') - name2 + 1;
  return (strcmp ((s1 ? s1 : name1 + i), (s2 ? s2 : name2 + j)));
}

void
maildir_getflag(char *name, int *d, int *f, int *r ,int *s, int *t)
{
  char tmp[MAILTMPLEN], *b;
  int offset = 0;

  if(d && f && r && s && t) 
    *d = *f = *r = *s = *t = 0;
  else
    return; 		/* can not call this function with null arguments */

  strcpy(tmp,name);
  while (b = strchr(tmp+offset, FLAGSEP)){
    char flag,last;
    int  i,k;
    if (!++b) break;
    switch (*b){
	case '1':
	case '2':
	case '3': flag = *b; b += 2;
		  for (k = 0; b[k] && b[k] != FLAGSEP && b[k] != ','; k++);
		  last = b[k];
		  b[k] = '\0';
		  if (flag == '2' || flag == '3'){
		     *d = strchr (b, MDFLAGC(Draft))   ? T : NIL;
		     *f = strchr (b, MDFLAGC(Flagged)) ? T : NIL;
		     *r = strchr (b, MDFLAGC(Replied)) ? T : NIL;
		     *s = strchr (b, MDFLAGC(Seen))    ? T : NIL;
		     *t = strchr (b, MDFLAGC(Trashed)) ? T : NIL;
		  }
		  b[k] = last;
		  b += k;
		  for (; tmp[offset] && tmp[offset] != FLAGSEP; offset++);
		  offset++;
		break;
	default: break;	/* Should we crash?... Nahhh */
    }
  }
}

int
maildir_message_in_list(char *msgname, struct direct **names, 
		unsigned long bottom, unsigned long top, unsigned long *pos)
{
  unsigned long middle = (bottom + top)/2;
  int test;

  if (!msgname)
     return NIL;

  if (pos) *pos = middle;

  if (same_maildir_file(msgname, names[middle]->d_name))
     return T;

  if (middle == bottom){	 /* 0 <= 0 < 1 */
     int rv = NIL;
     if (same_maildir_file(msgname, names[middle]->d_name)){
	rv = T;
	if (pos) *pos = middle;
     } 
     else
       if (same_maildir_file(msgname, names[top]->d_name)){
	  rv = T;
	  if (pos) *pos = top;
       }
     return rv;
  }

  test = comp_maildir_file(msgname, names[middle]->d_name);

  if (top <= bottom)
      return test ? NIL : T;

  if (test < 0 ) /* bottom <  msgname < middle */
     return maildir_message_in_list(msgname, names, bottom, middle, pos);
  else if (test > 0)  /* middle < msgname < top */
     return maildir_message_in_list(msgname, names, middle, top, pos);
  else return T;
}

void
maildir_abort(MAILSTREAM *stream)
{
  if (LOCAL){
    if (LOCAL->dir) fs_give ((void **) &LOCAL->dir);
    if (LOCAL->curdir) fs_give ((void **) &LOCAL->curdir);
    if (LOCAL->buf) fs_give ((void **) &LOCAL->buf);
    fs_give ((void **) &stream->local);
  }
  if (mdfpath) fs_give((void **)&mdfpath);
  stream->dtb = NIL;
}

int
maildir_contains_folder(char *dirname, char *name)
{
  char tmp[MAILTMPLEN], tmp2[MAILTMPLEN];
  int rv = 0;
  DIR *dir;
  struct direct *d;
  struct stat sbuf;
  char *maildirpath = mdirpath();

  maildir_file_path(dirname, tmp2);
  if(name){
    strcat(tmp2,"/");
    strcat(tmp2, name);
  }

  if (!(dir = opendir (tmp2)))
     return NIL;

  while (d = readdir(dir)){
    if (strcmp(d->d_name, ".") && strcmp(d->d_name,"..")
	&& strcmp(d->d_name, MDNAME(Cur)) 
	&& strcmp(d->d_name, MDNAME(Tmp)) 
	&& strcmp(d->d_name, MDNAME(New))){

       sprintf(tmp,"%s/%s", tmp2, d->d_name);
       if(maildir_valid(tmp)){
	  rv++;
	  break;
       }
    }
  }
  closedir(dir);
  return rv;
}

int
maildir_is_dir(char *dirname, char *name)
{
  char tmp[MAILTMPLEN];
  struct stat sbuf;
  char *maildirpath = mdirpath();

  maildir_file_path(dirname, tmp);
  if(name){
    strcat(tmp,"/");
    strcat(tmp,name);
  }
  strcat(tmp,"/");
  strcat(tmp,MDDIR);

  return ((stat(tmp, &sbuf) == 0) && S_ISREG (sbuf.st_mode)) ? 1 : 0;
}

int
maildir_dir_is_empty(char *mailbox)
{
  char tmp[MAILTMPLEN], tmp2[MAILTMPLEN];
  int rv = 1;
  DIR *dir;
  struct direct *d;
  struct stat sbuf;

  maildir_file_path(mailbox, tmp2);

  if (!(dir = opendir (tmp2)))
     return rv;

  while (d = readdir(dir)){
    sprintf(tmp,"%s/%s", tmp2, d->d_name);
    if (strcmp(d->d_name, ".") 
	&& strcmp(d->d_name,"..")
	&& strcmp(d->d_name, MDNAME(Cur)) 
	&& strcmp(d->d_name, MDNAME(Tmp)) 
	&& strcmp(d->d_name, MDNAME(New))
	&& strcmp(d->d_name, MDDIR)
	&& strcmp(d->d_name, MDUIDVALIDITY)
	&& !(d->d_name[0] == '.' 
		&& stat (tmp,&sbuf) == 0 
		&& S_ISREG(sbuf.st_mode))){
	   rv = 0;
	   break;
       }
  }
  closedir(dir);
  return rv;
}

void
maildir_get_file (MAILDIRFILE **mdfile)
{
  MAILDIRFILE *md;

  md = (MAILDIRFILE *) fs_get(sizeof(MAILDIRFILE));
  memset(md, 0, sizeof(MAILDIRFILE));
  *mdfile = md;
}

void
maildir_free_file (void **mdfile)
{
  MAILDIRFILE *md = (mdfile && *mdfile) ? (MAILDIRFILE *) *mdfile : NULL;

  if (md){
     if (md->name) fs_give((void **)&md->name);
     fs_give((void **)&md);
  }
}

void
maildir_free_file_only (void **mdfile)
{
  MAILDIRFILE *md = (mdfile && *mdfile) ? (MAILDIRFILE *) *mdfile : NULL;

  if (md && md->name) 
     fs_give((void **)&md->name);
}

int
maildir_any_new_msgs(char *mailbox)
{
  char tmp[MAILTMPLEN];
  int rv = NIL;
  DIR *dir;
  struct direct *d;
  struct stat sbuf;

  MDFLD(tmp, mailbox, New);

  if (!(dir = opendir (tmp)))
     return rv;

  while (d = readdir(dir)){
    if (d->d_name[0] == '.')
	continue;
    rv = T;
    break;
  }
  closedir(dir);
  return rv;
}


void
maildir_get_date(MAILSTREAM *stream, unsigned long msgno)
{
  MESSAGECACHE *elt;
  struct tm *t;
  unsigned long ti;
  int i,k;

  elt = mail_elt (stream,msgno);
  if ((ti = mdfntoul(MDFILE(elt), NIL)) > 0L){ /* from phile.c */
     t = gmtime (&ti);
     i = t->tm_hour * 60 + t->tm_min;
     k = t->tm_yday;
     t = localtime(&ti);
     i = t->tm_hour * 60 + t->tm_min - i;
     if (k = t->tm_yday - k) i += ((k < 0) == (abs (k) == 1)) ? -24*60 : 24*60;
     k = abs (i);
     elt->hours = t->tm_hour; 
     elt->minutes = t->tm_min; 
     elt->seconds = t->tm_sec;
     elt->day = t->tm_mday; elt->month = t->tm_mon + 1;
     elt->year = t->tm_year - (BASEYEAR - 1900);
     elt->zoccident = (k == i) ? 0 : 1;
     elt->zhours = k/60;
     elt->zminutes = k % 60;
  }
}

/* Support for Courier Style directories 

   When this code is complete there will be two types of support, which 
   will be configurable. The problem is the following: In Courier style 
   folder structure, a "folder" may have a subfolder called 
   "folder.subfolder", which is not natural in the file system in the 
   sense that I can not stat for "folder.subfolder" wihtout knowing what 
   "subfolder" is. It needs to be guessed. Because of this I need to look 
   in the list of folders if there is a folder with a name 
   "folder.subfolder", before I can say if the folder is dual or not. One 
   can avoid this annoyance if one ignores the problem by declaring that 
   every folder is dual. I will however code as the default the more 
   complicated idea of scaning the containing directory each time it is 
   modified and search for subfolders, and list the entries it found.
 */

int courier_dir_select (struct direct *name)
{
 return name->d_name[0] == '.' && (strlen(name->d_name) > 2
	|| (strlen(name->d_name) == 2 &&  name->d_name[1] != '.'));
}

int courier_dir_sort (const void *d1, const  void *d2)
{
  const struct direct **e1, **e2;

  e1 = (const struct direct **)d1;
  e2 = (const struct direct **)d2;

  return strcmp((char*)(*e1)->d_name, (char *)(*e2)->d_name);
}

void courier_free_cdir (COURIER_S **cdir)
{
  int i;

  if (!*cdir)
     return;

  if ((*cdir)->path) fs_give((void **)&((*cdir)->path));
  for (i = 0; i < (*cdir)->total; i++)
    if((*cdir)->data[i]->name) fs_give((void **)&((*cdir)->data[i]->name));
  fs_give((void **)&((*cdir)->data));
  fs_give((void **)&(*cdir));
}

COURIER_S *courier_get_cdir (int total)
{
 COURIER_S *cdir;

 cdir = (COURIER_S *)fs_get(sizeof(COURIER_S));
 memset(cdir, 0, sizeof(COURIER_S));
 cdir->data = (COURIERLOCAL **) fs_get(total*sizeof(COURIERLOCAL *));
 memset(cdir->data, 0, sizeof(COURIERLOCAL *));
 cdir->total = total;
 return cdir;
}

int courier_search_list(COURIERLOCAL **data, char *name, int first, int last)
{
  int try = (first + last)/2;

  if(!strstr(data[try]->name, name)){
     if(first == try) /* first == last || first + 1 == last */
	return strstr(data[last]->name, name) ? 1 : 0;
     if(strcmp(data[try]->name, name) < 0) /*data[try] < name < data[end] */
	return courier_search_list(data, name, try, last);
     else	/* data[begin] < name < data[try] */
	return courier_search_list(data, name, first, try);
  }
  return 1;
}

/* Lists all directories that are subdirectories of a given directory */

COURIER_S *courier_list_dir(char *curdir)
{
  struct direct **names = NIL;
  struct stat sbuf;
  unsigned long ndir;
  COURIER_S *cdir = NULL;
  char tmp[MAILTMPLEN], tmp2[MAILTMPLEN], pathname[MAILTMPLEN], 
	realname[MAILTMPLEN];
  char *maildirpath = mdirpath();
  int i, j, scand, td;

  /* There are two cases, either curdir is 
 	 #mc/INBOX.	 #mc/INBOX.foo
	or
	 #mc/Maildir/. 	 #mc/Maildir/.foo
   */
  strcpy(tmp,curdir + 4);
  if(!strncmp(ucase(tmp), "INBOX", 5))
    strcpy(tmp, "#mc/INBOX.");
  else{
   strcpy(tmp, curdir);
   for (i = strlen(tmp) - 1; tmp[i] && tmp[i] != '/'; i--);
   tmp[i+2] = '\0'; 	/* keep the last "." intact */
  }
  maildir_file_path(tmp, realname);
  maildir_scandir (realname, &names, &ndir, &scand, COURIER);

  if (scand > 0){
     cdir = courier_get_cdir(ndir);
     cdir->path = cpystr(realname);
     for(i = 0, j = 0; i < ndir; i++){
        td = realname[strlen(realname) - 1] == '.'
		&& *names[i]->d_name == '.';
	sprintf(tmp2,"%s%s", tmp, names[i]->d_name+1);
	sprintf(pathname,"%s%s", realname, names[i]->d_name + td);
	if(stat(pathname, &sbuf) == 0 && S_ISDIR(sbuf.st_mode)){
	   cdir->data[j] = (COURIERLOCAL *) fs_get(sizeof(COURIERLOCAL));
	   cdir->data[j++]->name = cpystr(tmp2);
	}
	fs_give((void **)&names[i]);
     }
     if(names)
       fs_give((void **) &names);
     cdir->total = j;
     if(cdir->total == 0)
        courier_free_cdir(&cdir);
  }
  return cdir;
}

void
courier_list_info(COURIER_S **cdirp, char *data, int i)
{
   long style = (long) maildir_parameters(GET_COURIERSTYLE, NIL);
   COURIER_S *cdir = *cdirp;

   if(maildir_valid(cdir->data[i]->name)){
      if(courier_search_list(cdir->data, data, 0, cdir->total - 1))
	 cdir->data[i]->attribute = LATT_HASCHILDREN;
      else
	 cdir->data[i]->attribute = (style == COURIER)
				? LATT_HASNOCHILDREN : LATT_NOINFERIORS;
   }
   else
      cdir->data[i]->attribute = LATT_NOSELECT;
      cdir->data[i]->attribute += maildir_any_new_msgs(cdir->data[i]->name) 
					? LATT_MARKED : LATT_UNMARKED;
}
