#if	!defined(lint) && !defined(DOS)
static char rcsid[] = "$Id: word.c 13655 2004-05-07 21:45:04Z jpf $";
#endif
/*
 * Program:	Word at a time routines
 *
 *
 * Michael Seibel
 * Networks and Distributed Computing
 * Computing and Communications
 * University of Washington
 * Administration Builiding, AG-44
 * Seattle, Washington, 98195, USA
 * Internet: mikes@cac.washington.edu
 *
 * Please address all bugs and comments to "pine-bugs@cac.washington.edu"
 *
 *
 * Pine and Pico are registered trademarks of the University of Washington.
 * No commercial use of these trademarks may be made without prior written
 * permission of the University of Washington.
 * 
 * Pine, Pico, and Pilot software and its included text are Copyright
 * 1989-2004 by the University of Washington.
 * 
 * The full text of our legal notices is contained in the file called
 * CPYRIGHT, included with this distribution.
 *
 */
/*
 * The routines in this file implement commands that work word at a time.
 * There are all sorts of word mode commands. If I do any sentence and/or
 * paragraph mode commands, they are likely to be put in this file.
 */

#include	"headers.h"


/* Word wrap on n-spaces. Back-over whatever precedes the point on the current
 * line and stop on the first word-break or the beginning of the line. If we
 * reach the beginning of the line, jump back to the end of the word and start
 * a new line.  Otherwise, break the line at the word-break, eat it, and jump
 * back to the end of the word.
 * Returns TRUE on success, FALSE on errors.
 */
wrapword()
{
    register int cnt;			/* size of word wrapped to next line */
    register int bp;			/* index to wrap on */
    register int first = -1;
    register int i;

    if(curwp->w_doto <= 0)		/* no line to wrap? */
      return(FALSE);

    for(bp = cnt = i = 0; cnt < llength(curwp->w_dotp) && !bp; cnt++, i++){
	if(isspace((unsigned char) lgetc(curwp->w_dotp, cnt).c)){
	    first = 0;
	    if(lgetc(curwp->w_dotp, cnt).c == TAB)
	      while(i+1 & 0x07)
		i++;
	}
	else if(!first)
	  first = cnt;

	if(first > 0 && i >= fillcol)
	  bp = first;
    }

    if(!bp)
      return(FALSE);

    /* bp now points to the first character of the next line */
    cnt = curwp->w_doto - bp;
    curwp->w_doto = bp;

    if(!lnewline())			/* break the line */
      return(FALSE);

    /*
     * if there's a line below, it doesn't start with whitespace 
     * and there's room for this line...
     */
    if(!(curbp->b_flag & BFWRAPOPEN)
       && lforw(curwp->w_dotp) != curbp->b_linep 
       && llength(lforw(curwp->w_dotp)) 
       && !isspace((unsigned char) lgetc(lforw(curwp->w_dotp), 0).c)
       && (llength(curwp->w_dotp) + llength(lforw(curwp->w_dotp)) < fillcol)){
	gotoeol(0, 1);			/* then pull text up from below */
	if(lgetc(curwp->w_dotp, curwp->w_doto - 1).c != ' ')
	  linsert(1, ' ');

	forwdel(0, 1);
	gotobol(0, 1);
    }

    curbp->b_flag &= ~BFWRAPOPEN;	/* don't open new line next wrap */
					/* restore dot (account for NL)  */
    if(cnt && !forwchar(0, cnt < 0 ? cnt-1 : cnt))
      return(FALSE);

    return(TRUE);
}


/*
 * Move the cursor backward by "n" words. All of the details of motion are
 * performed by the "backchar" and "forwchar" routines. Error if you try to
 * move beyond the buffers.
 */
backword(f, n)
    int f, n;
{
        if (n < 0)
                return (forwword(f, -n));
        if (backchar(FALSE, 1) == FALSE)
                return (FALSE);
        while (n--) {
                while (inword() == FALSE) {
                        if (backchar(FALSE, 1) == FALSE)
                                return (FALSE);
                }
                while (inword() != FALSE) {
                        if (backchar(FALSE, 1) == FALSE)
                                return (FALSE);
                }
        }
        return (forwchar(FALSE, 1));
}

/*
 * Move the cursor forward by the specified number of words. All of the motion
 * is done by "forwchar". Error if you try and move beyond the buffer's end.
 */
forwword(f, n)
    int f, n;
{
        if (n < 0)
                return (backword(f, -n));
        while (n--) {
#if	NFWORD
                while (inword() != FALSE) {
                        if (forwchar(FALSE, 1) == FALSE)
                                return (FALSE);
                }
#endif
                while (inword() == FALSE) {
                        if (forwchar(FALSE, 1) == FALSE)
                                return (FALSE);
                }
#if	NFWORD == 0
                while (inword() != FALSE) {
                        if (forwchar(FALSE, 1) == FALSE)
                                return (FALSE);
                }
#endif
        }
	return(TRUE);
}

#ifdef	MAYBELATER
/*
 * Move the cursor forward by the specified number of words. As you move,
 * convert any characters to upper case. Error if you try and move beyond the
 * end of the buffer. Bound to "M-U".
 */
upperword(f, n)
{
        register int    c;
	CELL            ac;

	ac.a = 0;
	if (curbp->b_mode&MDVIEW)	/* don't allow this command if	*/
		return(rdonly());	/* we are in read only mode	*/
        if (n < 0)
                return (FALSE);
        while (n--) {
                while (inword() == FALSE) {
                        if (forwchar(FALSE, 1) == FALSE)
                                return (FALSE);
                }
                while (inword() != FALSE) {
                        c = lgetc(curwp->w_dotp, curwp->w_doto).c;
                        if (c>='a' && c<='z') {
                                ac.c = (c -= 'a'-'A');
                                lputc(curwp->w_dotp, curwp->w_doto, ac);
                                lchange(WFHARD);
                        }
                        if (forwchar(FALSE, 1) == FALSE)
                                return (FALSE);
                }
        }
        return (TRUE);
}

/*
 * Move the cursor forward by the specified number of words. As you move
 * convert characters to lower case. Error if you try and move over the end of
 * the buffer. Bound to "M-L".
 */
lowerword(f, n)
{
        register int    c;
	CELL            ac;

	ac.a = 0;
	if (curbp->b_mode&MDVIEW)	/* don't allow this command if	*/
		return(rdonly());	/* we are in read only mode	*/
        if (n < 0)
                return (FALSE);
        while (n--) {
                while (inword() == FALSE) {
                        if (forwchar(FALSE, 1) == FALSE)
                                return (FALSE);
                }
                while (inword() != FALSE) {
                        c = lgetc(curwp->w_dotp, curwp->w_doto).c;
                        if (c>='A' && c<='Z') {
                                ac.c (c += 'a'-'A');
                                lputc(curwp->w_dotp, curwp->w_doto, ac);
                                lchange(WFHARD);
                        }
                        if (forwchar(FALSE, 1) == FALSE)
                                return (FALSE);
                }
        }
        return (TRUE);
}

/*
 * Move the cursor forward by the specified number of words. As you move
 * convert the first character of the word to upper case, and subsequent
 * characters to lower case. Error if you try and move past the end of the
 * buffer. Bound to "M-C".
 */
capword(f, n)
{
        register int    c;
	CELL	        ac;

	ac.a = 0;
	if (curbp->b_mode&MDVIEW)	/* don't allow this command if	*/
		return(rdonly());	/* we are in read only mode	*/
        if (n < 0)
                return (FALSE);
        while (n--) {
                while (inword() == FALSE) {
                        if (forwchar(FALSE, 1) == FALSE)
                                return (FALSE);
                }
                if (inword() != FALSE) {
                        c = lgetc(curwp->w_dotp, curwp->w_doto).c;
                        if (c>='a' && c<='z') {
			    ac.c = (c -= 'a'-'A');
			    lputc(curwp->w_dotp, curwp->w_doto, ac);
			    lchange(WFHARD);
                        }
                        if (forwchar(FALSE, 1) == FALSE)
                                return (FALSE);
                        while (inword() != FALSE) {
                                c = lgetc(curwp->w_dotp, curwp->w_doto).c;
                                if (c>='A' && c<='Z') {
				    ac.c = (c += 'a'-'A');
				    lputc(curwp->w_dotp, curwp->w_doto, ac);
				    lchange(WFHARD);
                                }
                                if (forwchar(FALSE, 1) == FALSE)
                                        return (FALSE);
                        }
                }
        }
        return (TRUE);
}

/*
 * Kill forward by "n" words. Remember the location of dot. Move forward by
 * the right number of words. Put dot back where it was and issue the kill
 * command for the right number of characters. Bound to "M-D".
 */
delfword(f, n)
{
        register long   size;
        register LINE   *dotp;
        register int    doto;

	if (curbp->b_mode&MDVIEW)	/* don't allow this command if	*/
		return(rdonly());	/* we are in read only mode	*/
        if (n < 0)
                return (FALSE);
        dotp = curwp->w_dotp;
        doto = curwp->w_doto;
        size = 0L;
        while (n--) {
#if	NFWORD
		while (inword() != FALSE) {
			if (forwchar(FALSE,1) == FALSE)
				return(FALSE);
			++size;
		}
#endif
                while (inword() == FALSE) {
                        if (forwchar(FALSE, 1) == FALSE)
                                return (FALSE);
                        ++size;
                }
#if	NFWORD == 0
                while (inword() != FALSE) {
                        if (forwchar(FALSE, 1) == FALSE)
                                return (FALSE);
                        ++size;
                }
#endif
        }
        curwp->w_dotp = dotp;
        curwp->w_doto = doto;
        return (ldelete(size, kinsert));
}

/*
 * Kill backwards by "n" words. Move backwards by the desired number of words,
 * counting the characters. When dot is finally moved to its resting place,
 * fire off the kill command. Bound to "M-Rubout" and to "M-Backspace".
 */
delbword(f, n)
{
        register long   size;

	if (curbp->b_mode&MDVIEW)	/* don't allow this command if	*/
		return(rdonly());	/* we are in read only mode	*/
        if (n < 0)
                return (FALSE);
        if (backchar(FALSE, 1) == FALSE)
                return (FALSE);
        size = 0L;
        while (n--) {
                while (inword() == FALSE) {
                        if (backchar(FALSE, 1) == FALSE)
                                return (FALSE);
                        ++size;
                }
                while (inword() != FALSE) {
                        if (backchar(FALSE, 1) == FALSE)
                                return (FALSE);
                        ++size;
                }
        }
        if (forwchar(FALSE, 1) == FALSE)
                return (FALSE);
        return (ldelete(size, kinsert));
}
#endif	/* MAYBELATER */

/*
 * Return TRUE if the character at dot is a character that is considered to be
 * part of a word. The word character list is hard coded. Should be setable.
 */
inword()
{
    return(curwp->w_doto < llength(curwp->w_dotp)
	   && isalnum((unsigned char)lgetc(curwp->w_dotp, curwp->w_doto).c));
}

/* Support of indentation of paragraphs */
#define UCH(c) ((unsigned char) (c))
#define NBSP UCH('\240')
#define ISspace(c) (UCH(c) == ' ' || UCH(c) == TAB || UCH(c) == NBSP)
#define is_indent_char(c)  (((c) == '.' || (c) == '}' || (c) == RPAREN || \
			     (c) == '*' || (c) == '+' || is_a_digit(c) || \
			     ISspace(c) || (c) == '-' || \
			     (c) == ']') ? 1 : 0)
#define allowed_after_digit(c,word,k)  ((((c) == '.' && \
				   allowed_after_period(next((word),(k))))  ||\
				  (c) == RPAREN || (c) == '}' || (c) == ']' ||\
				  ISspace(c) ||  is_a_digit(c) || \
				  ((c) == '-' ) && \
					allowed_after_dash(next((word),(k)))) \
				  ? 1 : 0)
#define allowed_after_period(c)  (((c) == RPAREN || (c) == '}' || (c) == ']' ||\
				   ISspace(c) || (c) == '-' || \
				   is_a_digit(c)) ? 1 : 0)
#define allowed_after_parenth(c)  (ISspace(c) ? 1 : 0)
#define allowed_after_space(c)    (ISspace(c) ? 1 : 0)
#define allowed_after_braces(c)   (ISspace(c) ? 1 : 0)
#define allowed_after_star(c)     ((ISspace(c) || (c) == RPAREN ||\
					(c) == ']' || (c) == '}') ? 1 : 0)
#define allowed_after_dash(c)     ((ISspace(c) || is_a_digit(c)) ? 1 : 0)
#define EOLchar(c)		  (((c) == '.' || (c) == ':' || (c) == '?' ||\
				   (c) == '!') ? 1 : 0)

int is_indent PROTO((char *, int));
int indent_match PROTO(( char *, LINE *, char *, int, int));
int get_indent_raw_line PROTO((char *, char *, char *, int, int, int));

/* Extended justification support */

#define is_cquote(c) ((c) == '>' || (c) == '|' || (c) == ']' || (c) == ':')
#define is_cword(c)  ((((c) >= 'a') && ((c) <= 'z')) ||  \
		      (((c) >= 'A') && ((c) <= 'Z')) || \
		      (((c) >= '0') && ((c) <= '9')) || \
		       ((c) == ' ') || ((c) == '?') || \
		       ((c) == '@') || ((c) == '.') || \
		       ((c) == '!') || ((c) == '\'') || \
		       ((c) == ',') || ((c) == '\"') ? 1 : 0)
#define isaquote(c)   ((c) == '\"' || (c) == '\'')
#define is8bit(c)	((((int) (c)) & 0x80) ? 1 : 0)
#define forbidden(c)	(((c) == '\"') || ((c) == '\'') || ((c) == '$') ||\
			((c) == ',')  || ((c) == '.')  || ((c) == '-') ||\
			((c) == LPAREN) || ((c) == '/')|| ((c) == '`') ||\
			((c) == '{') || ((c) == '\\') || \
			(((c) >= '0')  && ((c) <= '9')))
#define is_cletter(c)  ((((c) >= 'a') && ((c) <= 'z'))) ||\
			 ((((c) >= 'A') && ((c) <= 'Z'))||\
			is8bit(c))
#define is_cnumber(c) ((c) >= '0' && (c) <= '9')
#define allwd_after_word(c) (((c) == ' ') || ((c) == '>') || is_cletter(c))
#define allwd_after_qsword(c)  (((c) != '\\') && ((c) != RPAREN))
#define before(word,i) (((i) > 0) ? (word)[(i) - 1] : 0)
#define next(w,i) ((((w)[(i)]) != 0) ? ((w)[(i) + 1]) : 0)
#define now(w,i)  ((w)[(i)])
#define is_qsword(c)  (((c) == ':') || ((c) == RPAREN) ? 1 : 0)
#define is_colon(c)   (((c) == ':') ? 1 : 0)
#define is_rarrow(c)  (((c) == '>') ? 1 : 0)
#define is_tilde(c)   (((c) == '~') ? 1 : 0)
#define is_dash(c)    (((c) == '-') ? 1 : 0)
#define is_pound(c)   (((c) == '#') ? 1 : 0)
#define is_space(c)   (((c) == ' ') ? 1 : 0)
#define is_a_digit(c) ((((c) >= '0') && ((c) <= '9')) ? 1 : 0)
#define is_allowed(c)  (is_cquote(c) || is_cword(c) || is_dash(c) || \
			is_pound(c))

/* Internal justification functions */

QSTRING_S *is_quote PROTO((char *, char *, int));
QSTRING_S *copy_qs PROTO((QSTRING_S *));
QSTRING_S *qs_normal_part PROTO((QSTRING_S *));
QSTRING_S *qs_remove_trailing_spaces PROTO((QSTRING_S *));
QSTRING_S *trim_qs_from_cl PROTO((QSTRING_S *, QSTRING_S *, QSTRING_S *));
QSTRING_S *fix_qstring PROTO((QSTRING_S *, QSTRING_S *, QSTRING_S *));
QSTRING_S *qs_add PROTO((char *, char *, QStrType, int, int, int, int));
QSTRING_S *remove_qsword PROTO((QSTRING_S *));
QSTRING_S *qs_quote_match PROTO((char *, LINE *, char *, int));
int	   qstring_is_normal PROTO((QSTRING_S *));
int	   exists_good_part PROTO((QSTRING_S *));
int	   value_is_space PROTO((char *));
int	   strcmp_qs PROTO((char *, char *));
int	   count_levels_qstring PROTO((QSTRING_S *));
int	   same_qstring PROTO((QSTRING_S *, QSTRING_S *));
int	   advance_quote_string PROTO((char *, char *, int));
int	   strlenis PROTO((char *));
void	   linencpy PROTO((char *, LINE *, int));

char *
default_qstr()
{
 return (glo_quote_str ? glo_quote_str
	: (Pmaster && Pmaster->quote_str) ? Pmaster->quote_str : "");
}

/* 
 * This function creates a qstring pointer with the information that 
 * is_quote handles to it.
 * Parameters:
 * qs		- User supplied quote string
 * word		- The line of text that the user is trying to read/justify
 * beginw	- Where we need to start copying from
 * endw		- Where we end copying
 * offset	- Any offset in endw that we need to account for
 * typeqs	- type of the string to be created
 * neednext	- boolean, indicating if we need to compute the next field
 *		  of leave it NULL
 *
 * It is a mistake to call this function if beginw >= endw + offset.
 * Please note the equality sign in the above inequality (this is because
 * we always assume that qstring->value != "").
 */

QSTRING_S *
qs_add(qs, word, typeqs, beginw, endw, offset, neednext)
   char	     *qs;
   char	      word[NSTRING];
   QStrType   typeqs;
   int	      beginw;
   int	      endw;
   int	      offset;
   int	      neednext;
{
    QSTRING_S *qstring, *nextqs = (QSTRING_S *) NULL;
    int i;

    qstring = (QSTRING_S *) malloc (sizeof(QSTRING_S));
    memset (qstring, 0, sizeof(QSTRING_S));
    qstring->qstype = qsNormal;

    if (beginw == 0){
	beginw = endw + offset;
	qstring->qstype = typeqs;
    }

    if (neednext)
       nextqs = is_quote(qs, word+beginw, 1);

    qstring->value = (char *) malloc((beginw+1)*sizeof(char));
    strncpy(qstring->value, word, beginw);
    qstring->value[beginw] = '\0';

    qstring->next = nextqs;

    return qstring;
}


int
qstring_is_normal(cl)
  QSTRING_S *cl;
{
  for (;cl && (cl->qstype == qsNormal); cl = cl->next);

  return cl ? 0 : 1;
}

void
free_qs(cl)
  QSTRING_S **cl;
{
  if (!(*cl))
    return;

  if ((*cl)->next)
    free_qs(&((*cl)->next));

  (*cl)->next = (QSTRING_S *) NULL;

  if ((*cl)->value)
    free((void *)(*cl)->value);

  (*cl)->value = (char *) NULL;

  free((void *)(*cl));
  *cl = (QSTRING_S *) NULL;
}

QSTRING_S *
copy_qs(cl)
  QSTRING_S *cl;
{
  QSTRING_S *qs;

  if (!cl)
    return (QSTRING_S *)NULL;

  qs = (QSTRING_S *) malloc (sizeof(QSTRING_S));
  memset (qs, 0, sizeof(QSTRING_S));

  qs->value = (char *) malloc ((strlen(cl->value)+1)*sizeof(char));
  strcpy(qs->value, cl->value);
  qs->qstype = cl->qstype;
  qs->next = copy_qs(cl->next);
  return qs;
}

/* 
 * Given a quote string, this function returns the part that is the leading
 * normal part of it. (the normal part is the part that is tagged qsNormal,
 * that is to say, the one that is not controversial at all (like qsString
 * for example).
 */
QSTRING_S *
qs_normal_part(cl)
  QSTRING_S *cl;
{

  if (!cl)		/* nothing in, nothing out */
     return cl;

  if (cl->qstype != qsNormal)
     free_qs(&cl);

  if (cl) 
     cl->next = qs_normal_part(cl->next);

  return cl;
}

int
value_is_space(value)
  char *value;
{
  for (; value && *value && ISspace(*value); value++);

  return value && *value ? 0 : 1;
}

/* 
 * this function removes trailing spaces from a quote string, but leaves the
 * last one if there are trailing spaces
 */
QSTRING_S *
qs_remove_trailing_spaces(cl)
   QSTRING_S *cl;
{
  QSTRING_S *rl = cl;

  if (!cl)		/* nothing in, nothing out */
    return cl;

  if (cl->next)
    cl->next = qs_remove_trailing_spaces(cl->next);
  else{
    if (value_is_space(cl->value))
       free_qs(&cl);
    else{
       int i, l; 
       i = l = strlen(cl->value) - 1;
       while (cl->value && cl->value[i] 
	  && ISspace(cl->value[i]))
	     i--;
	  i += (i < l) ? 2 : 1;
	  cl->value[i] = '\0';
    }
  }

  return cl;
}

/*
 * This function returns if two strings are the same quote string.
 * The call is not symmetric. cl must preceed the line nl. This function
 * should be called for comparing the last part of cl and nl.
 */
int
strcmp_qs(valuecl, valuenl)
  char *valuecl;
  char *valuenl;
{
   int j;

   for (j = 0; valuecl[j] && (valuecl[j] == valuenl[j]); j++);
   return !strcmp(valuecl, valuenl)
	    || (valuenl[j] && value_is_space(valuenl+j) 
			   && value_is_space(valuecl+j)
			   && strlenis(valuecl+j) >= strlenis(valuenl+j))
	    || (!valuenl[j] && value_is_space(valuecl+j));
}

int
count_levels_qstring(cl)
  QSTRING_S *cl;
{
  int count;
  for (count = 0; cl ; count++, cl = cl->next);

  return count;
}

/* 
 * This function returns the number of agreements between 
 * cl and nl. The call is not symmetric. cl must be the line
 * preceding nl.
 */
int
same_qstring(cl,nl)
  QSTRING_S *cl;
  QSTRING_S *nl;
{
   int same = 0, done = 0;

   for (;cl && nl && !done; cl = cl->next, nl = nl->next)
      if ((cl->qstype == nl->qstype) && (!strcmp(cl->value, nl->value) 
	  || ((!cl->next) && strcmp_qs(cl->value, nl->value))))
	same++;
      else
	done++;

   return same;
}

QSTRING_S *
trim_qs_from_cl(cl, nl, pl)
  QSTRING_S *cl;
  QSTRING_S *nl;
  QSTRING_S *pl;
{
   QSTRING_S *cqstring = pl ? pl : nl;
   QSTRING_S *tl = pl ? pl : nl;
   int p, c;

   if (qstring_is_normal(tl))
    return tl;

    p = same_qstring(pl ? pl : cl, pl ? cl : nl);

    for (c = 1; c < p; c++, cl = cl->next, tl = tl->next);

    /*
     * cl->next and tl->next differ, it may be because cl->next does not
     * exist or tl->next does not exist or simply both exist but are
     * different. In this last case, it may be that cl->next->value is made
     * of spaces. If this is the case, tl advances once more.
     */

    if (tl->next){
       if (cl && cl->next && value_is_space(cl->next->value))
	  tl = tl->next;
       if (tl->next)
	  free_qs(&(tl->next));
    }

    if (!p)
       free_qs(&cqstring);

    return cqstring;
}

/* This function trims cl so that it returns a real quote string based
 * on information gathered from the previous and next lines. pl and cl are
 * also trimmed, but that is done in another function, not here.
 */
QSTRING_S *
fix_qstring(cl, nl, pl)
  QSTRING_S *cl;
  QSTRING_S *nl;
  QSTRING_S *pl;
{
  QSTRING_S *cqstring = cl, *nqstring = nl, *pqstring = pl;
  int c, n;

  if (qstring_is_normal(cl))
    return cl;

  c = count_levels_qstring(cl);
  n = same_qstring(cl,nl);

  if (!n){  /* no next line or no agreement with next line */
     int p = same_qstring(pl, cl); /* number of agreements between pl and cl */
     QSTRING_S *tl; 		   /* test line */

     /*
      * Here p <= c, so either p < c or p == c. If p == c, we are done, 
      * and return cl. If not, there are two cases, either p == 0 or
      * 0 < p < c. In the first case, we do not have enough evidence
      * to return anything other than the normal part of cl, in the second
      * case we can only return p levels of cl.
      */

     if (p == c)
	tl = cqstring;
     else{
	if (p){
	   for (c = 1; c < p; c++)
	      cl = cl->next;
	   free_qs(&(cl->next));
	   tl = cqstring;
	}
	else{
	   int done = 0;
	   QSTRING_S *al = cl;	/* another line */
	   /* 
	    * Ok, we reaelly don't have enough evidence to return anything,
	    * different from the normal part of cl, but it could be possible
	    * that we may want to accept the not-normal part, so we better
	    * make an extra test to determine what needs to be freed
	    */
	   while (pl && cl && !strucmp(cl->value, pl->value)){
		cl = cl->next;
		pl = pl->next;
	   }
	   if (pl && cl && strcmp_qs(pl->value, cl->value))
		cl = cl->next;	/* next level differs only in spaces */
	   while (!done){
		while (cl && cl->qstype == qsNormal)
		    cl = cl->next;
		if (cl){
		   if ((cl->qstype == qsString) 
		       && (cl->value[strlen(cl->value) - 1] == '>'))
		      cl = cl->next;
		   else done++;
		}
		else done++;
	   }
	   if (al == cl){
	      free_qs(&(cl));
	      tl = cl;
	   }
	   else {
	      while (al && (al->next != cl))
	         al = al->next;
	      cl = al;
	      if (cl && cl->next)
	         free_qs(&(cl->next));
	      tl = cqstring;
	   }
	}
     }
     return tl;
  }
  if (n + 1 < c){  /* if there are not enough agreements */
     int p = same_qstring(pl, cl); /* number of agreement between pl and cl */
     QSTRING_S *tl; /* test line */

	/*
	 * There's no way we can use cl in this case, but we can use
	 * part of cl, this is if pl does not have more agreements
	 * with cl.
	 */

     if (p == c)
	tl = cqstring;
     else{
	int m = p < n ? n : p;
	for (c = 1; c < m; c++){
	  pl = pl ? pl->next : (QSTRING_S *) NULL;
	  nl = nl ? nl->next : (QSTRING_S *) NULL;
	  cl = cl->next;
	}
	if ((p == n) && pl && pl->next && nl && nl->next
	   && ((cl->next->qstype == pl->next->qstype) 
	      || (cl->next->qstype == nl->next->qstype))
	   && (strcmp_qs(cl->next->value, pl->next->value) 
	      || strcmp_qs(pl->next->value, cl->next->value) 
	      || strcmp_qs(cl->next->value, nl->next->value) 
	      || strcmp_qs(nl->next->value, cl->next->value))) 
	  cl = cl->next;	/* next level differs only in spaces */
	if (cl->next)
	   free_qs(&(cl->next));
	tl = cqstring;
     }
     return tl;
  }
  if (n + 1 == c){
     int p = same_qstring(pl, cl);
     QSTRING_S *tl; /* test line */

     /* 
      * p <= c, so p <= n+1, which means p < n + 1 or p == n + 1.
      * If p < n + 1, then p <= n.
      * so we have three possibilities: 
      *	p == n + 1 or p == n or p < n.
      * In the first case we copy p == n + 1 == c levels, in the second
      * and third case we copy n levels, and check if we can copy the 
      * n + 1 == c level.
      */

     if (p == n + 1)      /* p == c, in the above sense of c */
	tl = cl; 	  /* use cl, this is enough evidence */
     else{
	for (c = 1; c < n; c++)
	  cl = cl->next;
	/* 
	 * Here c == n, we only have one more level of cl, and at least one
	 * more level of nl 
	 */
	if (cl->next->qstype == qsNormal)
	   cl = cl->next;
	if (cl->next)
	   free_qs(&(cl->next));
	tl = cqstring;
     }
     return tl;
  }
  if (n == c)  /* Yeah!!! */
    return cqstring;
}

/* 
 * This function flattens the quote string returned to us by is_quote. A
 * crash in this function implies a bug elsewhere.
 */
void 
flatten_qstring(qs, buff, bufflen)
  QSTRING_S *qs;
  char	    *buff;
  int	     bufflen;
{
  int i, j;

  if(!buff || bufflen <= 0)
    return;

  for (i = 0; qs; qs = qs->next)
    for (j = 0; i < bufflen - 1
		&&  (qs->value[j]) && (buff[i++] = qs->value[j]); j++);
  buff[i] = '\0';
}

/* 
 * Given a string, we return the position where the function thinks that
 * the quote string is over, if you are ever thinking of fixing something,
 * you got to the right place. Memory freed by caller. Experience shows
 * that it only makes sense to initialize memory when we need it, not at
 * the start of this function.
 */
QSTRING_S *
is_quote (qs,word, been_here)
  char *qs;
  char word[NSTRING];
  int been_here;
{
  int i = 0, j, c, nxt, prev, finished = 0, offset;
  QSTRING_S *qstring = (QSTRING_S *) NULL;

   if (!word || !word[0])
      return (QSTRING_S *) NULL;

   while (!finished){
	/* 
	 * Before we apply our rules, let's advance past the quote string
	 * given by the user, this will avoid not recognition of the
	 * user's indent string and application of the arbitrary rules
	 * below. Notice that this step may bring bugs into this
	 * procedure, but these bugs will only appear if the indent string
	 * is really really strange and the text to be justified
	 * cooperates a lot too, so in general this will not be a problem.
	 * If you are concerned about this bug, simply remove the
	 * following lines after this comment and before the "switch"
	 * command below and use a more normal quote string!.
	 */
	i += advance_quote_string(qs, word, i);
	if (!word[i]) /* went too far? */
	   return qs_add(qs, word, qsNormal, 0, i, 0, 0);

     switch (c = now(word,i)){
	case NBSP:
	case TAB :
	case ' ' : { QSTRING_S *nextqs, *testqs = NULL;
		    int j;

		    for (; ISspace(word[i]); i++);
		    nextqs = is_quote(qs,word+i, 1);
		  /* 
		   * Merge qstring and nextqs, since this is an artificial
		   * separation, unless nextqs is of different type.
		   * What this means in practice is that if 
		   * qs->qstype == qsNormal and qs->next != NULL, then
		   * qs->next->qstype != qsNormal.
		   *
		   * Can't use qs_add to merge because it could lead
		   * to an infinite loop (e.g a line "^ ^").
		   */
		    if (nextqs){
		        if(nextqs->qstype == qsNormal){
		          i += strlen(nextqs->value);
		          testqs = copy_qs(nextqs->next);
		        }
		        else
		          testqs = copy_qs(nextqs);
		        free_qs(&nextqs);
		    }

		    qstring = (QSTRING_S *) malloc (sizeof(QSTRING_S));
		    memset (qstring, 0, sizeof(QSTRING_S));

		    qstring->value = (char *) malloc((i+1)*sizeof(char));
		    strncpy(qstring->value, word, i);
		    qstring->value[i] = '\0';
		    qstring->qstype   = qsNormal;
		    qstring->next     = testqs;

		    return qstring;
		  }
		 break;

	case RPAREN:		/* parenthesis ')' */
		     if ((i != 0) || ((i == 0) && been_here))
			i++;
		     else
			if (i == 0)
		           return qs_add(qs, word, qsChar, i, i, 1, 1);
		        else
			   finished++;
		    break;

	case ';':
	case ':':			/* colon */
	case '~': nxt = next(word,i);
		  if (is_tilde(c) && (nxt == '/'))
		     finished++;
		  else if (is_cquote(c)
			|| is_cquote(nxt)
			|| ((c != '~') && (nxt == RPAREN))
			|| ((i != 0) && is_space(nxt))
			|| is_cquote(prev = before(word,i)) 
			|| (is_space(prev) && !is_tilde(c))
			|| (is_tilde(c) && nxt != '/'))
		      i++;
		  else if (i == 0 && been_here)
		       return qs_add(qs, word, qsChar, i, i, 1, 1);
		  else
		       finished++;
		 break;

	case '<' :
	case '=' :
	case '-' : offset = is_cquote(nxt = next(word,i)) ? 2
                                : ((nxt == c)
                                     && is_cquote(next(word,i+1))) ? 3 : -1;
                   
		   if (offset > 0)
		       return qs_add(qs, word, qsString, i, i, offset, 1);
		   else
		       finished++;
		 break;

	case '[' :
	case '+' :	/* accept +>, *> */
	case '*' :  if (is_rarrow(nxt = next(word, i)) || /* stars */
		      (is_space(nxt) && is_rarrow(next(word,i+1))))
			 i++;
		    else
		       finished++;
		 break;

	case '^' :
	case '!' :
	case '%' :
	case '#' : if (next(word,i) != c)
		      return qs_add(qs, word, qsChar, i, i+1, 0, 1);
		   else
		      finished++;
	         break;

        default:
	    if (is_cquote(c))
	       i++;
	    else if (is_cletter(c)){
		for (j = i; (is_cletter(nxt = next(word,j)) || is_cnumber(nxt))
			    && !(is_space(nxt));j++);
		  /* 
		   * The whole reason why we are splitting the quote
		   * string is so that we will be able to accept quote
		   * strings that are strange in some way. Here we got to
		   * a point in which a quote string might exist, but it
		   * could be strange, so we need to create a "next" field
		   * for the quote string to warn us that something
		   * strange is coming. We need to confirm if this is a
		   * good choice later. For now we will let it pass.
		   */
		  if (isaword(word,i,j) || isamailbox(word,i,j)){
		    int offset;
		    QStrType qstype;

		    offset = (is_cquote(c = next(word,j))
			     || (c == RPAREN)) ? 2
				: ((is_space(c) 
				     && is_cquote(next(word,j+1))) ? 3 : -1);

		    qstype = (is_cquote(c) || (c == RPAREN))
		      ? (is_qsword(c) ? qsWord : qsString)
		      : ((is_space(c) && is_cquote(next(word,j+1)))
			 ? (is_qsword(next(word,j+1))
			    ? qsWord : qsString) 
				  : qsString);

		    /* 
		     * qsWords are valid quote strings only when
		     * they are followed by text.
		     */
		    if ((offset > 0) && (qstype == qsWord) &&
			!(allwd_after_qsword(now(word,j + offset))))
			offset = -1;

		    if (offset > 0)
		        return qs_add(qs, word, qstype, i, j, offset, 1);
		  }
		  finished++;
	    }
	    else{
		if(!forbidden(c))
		   return qs_add(qs, word, qsChar, 0, 1, 0, 1);
		else	/* chao pescao */
		   finished++;
	    }
	break;
     }  /* End Switch */
   }  /* End while */

   if (i > 0)
      qstring = qs_add(qs, word, qsNormal, 0, i, 0, 0);

   return qstring;
}

void
linencpy(word, l, buflen)
char word[NSTRING];
LINE *l;
int buflen;
{
  int i = 0;
  for (;(i < buflen) && (i < llength(l)) && (word[i] = (char)lgetc(l,i).c); i++);
  word[buflen - 1] = '\0';
}

int
isaword(word,i,j)
char word[NSTRING];
int i;
int j;
{
 return i <= j && is_cletter(word[i]) ?
         (i < j ? isaword(word,i+1,j) : 1) : 0;
}

int
isamailbox(word,i,j)
char word[NSTRING];
int i;
int j;
{
 return i <= j && (is_cletter(word[i]) || is_a_digit(word[i]) 
		  || word[i] == '.')
	? (i < j ? isamailbox(word,i+1,j) : 1) : 0;
}

/* 
 * This function returns the quote string as a structure. In this way we
   have two ways to get the quote string: as a char * or as a QSTRING_S *
   directly.
 */
QSTRING_S *
qs_quote_match(q, l, rqstr, rqstrlen)
    char *q;
    LINE *l;
    char *rqstr;
    int   rqstrlen;
{
    char GLine[NSTRING] = {'\0'}, NLine[NSTRING] = {'\0'}, 
	 PLine[NSTRING] = {'\0'};
    LINE *nl = l != curbp->b_linep ? lforw(l) : NULL;
    LINE *pl = lback(l) != curbp->b_linep ? lback(l) : NULL;
    int plb = 1;

    linencpy(GLine, l, NSTRING);

    if (nl)
      linencpy(NLine, nl, NSTRING);

    if (pl){
      linencpy(PLine, pl, NSTRING);
      if(lback(pl) != curbp->b_linep){
	char PPLine[NSTRING] = {'\0'};

	linencpy(PPLine, lback(pl), NSTRING);
	plb = line_isblank(q, PLine, GLine, PPLine, NSTRING);
      }
    }

    return do_quote_match(q, GLine, NLine, PLine, rqstr, rqstrlen, plb);
}

/*
 * Return number of quotes if whatever starts the line matches the quote
 * string.
 * rqstring is  pointer to raw qstring, buf points to processed qstring
 */
quote_match(q, l, buf, buflen, raw)
    char *q;
    LINE *l;
    char *buf;
    int   buflen;
    int   raw;
{
    QSTRING_S *qs;
    char rqstr[NSTRING] = {'\0'}; 

    qs = qs_quote_match(q, l, rqstr, NSTRING);
    flatten_qstring(qs, buf, buflen);
    if (qs)
       free_qs(&qs);

    if(raw){
      strncpy(buf, rqstr, buflen < NSTRING ? buflen : NSTRING);
      buf[buflen-1] = '\0';
    }

    return  buf && buf[0] ? strlen(buf) : 0;
}

/*
   This routine removes the last part that is qsword or qschar that is not
   followed by a normal part. This means that if a qsword or qschar is
   followed by a qsnormal (or qsstring), we accept the qsword (or qschar)
   as part of a quote string.
 */

QSTRING_S *
remove_qsword(cl)
  QSTRING_S *cl;
{
    QSTRING_S *np = cl;
    QSTRING_S *cp = np;		/* this variable trails cl */

    while(1){
       while (cl && cl->qstype == qsNormal)
           cl = cl->next;

       if (cl){
	  if (((cl->qstype == qsWord) || (cl->qstype == qsChar))
		&& !exists_good_part(cl)){
	      if (np == cl)     /* qsword or qschar at the beginning */
		 free_qs(&cp);
	      else{
	         while (np->next != cl)
		     np = np->next;
		 free_qs(&(np->next));
	     }
	     break;
	  }
	  else
	     cl = cl->next;
       }
       else
        break;
   }
   return cp;
}

int
exists_good_part (cl)
  QSTRING_S *cl;
{
  return (cl ? (((cl->qstype != qsWord) && (cl->qstype != qsChar)
                   && !value_is_space(cl->value)) 
		? 1 
		: exists_good_part(cl->next)) 
	    : 0);
}

line_isblank(q, GLine, NLine, PLine, buflen)
   char *q, *GLine, *PLine, *NLine;
   int buflen;
{
    int n = 0;
    QSTRING_S *cl;
    char qstr[NSTRING];

    cl = do_raw_quote_match(q, GLine, NLine, PLine, NULL, NULL);

    flatten_qstring(cl, qstr, NSTRING);

    free_qs(&cl);

    for(n = strlen(qstr); n < buflen && GLine[n]; n++)
      if(!isspace((unsigned char) GLine[n]))
        return(FALSE);

    return(TRUE);
}


QSTRING_S *
do_raw_quote_match(q, GLine, NLine, PLine, nlp, plp)
  char *q, *GLine, *NLine, *PLine;
  QSTRING_S **nlp, **plp;
{
  QSTRING_S *cl, *nl = NULL, *pl = NULL;
  char nbuf[NSTRING], pbuf[NSTRING], buf[NSTRING];
  int emptypl = 0, emptynl = 0;

    cl = is_quote(q, GLine, 0);	/* Current or Given line */

    if (!cl)			/* if nothing in, nothing out */
      return cl;

    if (NLine && NLine[0]){
	nl = is_quote(q, NLine, 0);	/* Next Line		 */
	if(nlp)
	  *nlp = nl;
    }
    if (PLine && PLine[0]){
	pl = is_quote(q, PLine, 0);	/* Previous Line	 */
	if(plp)
	  *plp = pl;
    }

    /* 
     * If there's nothing in the preceeding or following line
     * there is not enough information to accept it or discard it. In this
     * case it's likely to be an isolated line, so we better accept it
     * if it does not look like a word. */

    flatten_qstring(pl, pbuf, NSTRING);
    emptypl = (!PLine || !PLine[0] ||
		(pl && value_is_space(pbuf)) && !PLine[strlen(pbuf)]) ? 1 : 0;
    if (emptypl){
       flatten_qstring(nl, nbuf, NSTRING);
       emptynl = (!NLine || !NLine[0] ||
		(nl && value_is_space(nbuf) && !NLine[strlen(nbuf)])) ? 1 : 0;
       if (emptynl){
	  cl = remove_qsword(cl);
	  cl = qs_remove_trailing_spaces(cl);
	  free_qs(&nl);
	  free_qs(&pl);
	  if(nlp) *nlp = NULL;
	  if(plp) *plp = NULL;

	  return cl;
       }
    }

    /* 
     * If either cl, nl or pl contain suspicious characters that may make
     * them (or not) be quote strings, we need to fix them, so that the
     * next pass will be done correctly.
     */

    cl = fix_qstring(cl, nl, pl);
    nl = trim_qs_from_cl(cl, nl, NULL);
    pl = trim_qs_from_cl(cl, NULL, pl);

    if(nlp) *nlp = nl;
    if(plp) *plp = pl;

    return cl;
}

QSTRING_S *
do_quote_match(q, GLine, NLine, PLine, rqstr, rqstrlen, plb)
  char *q, *GLine, *NLine, *PLine, *rqstr;
  int   rqstrlen, plb;
{
    QSTRING_S *cl, *nl = NULL, *pl = NULL;
    int c, n, p,i, j, NewP, NewC, NewN, clength, same = 0;
    char nbuf[NSTRING], pbuf[NSTRING], buf[NSTRING];

    cl = do_raw_quote_match(q, GLine, NLine, PLine, &nl, &pl);

    if (!cl)			/* if nothing in, nothing out */
      return cl;

    flatten_qstring(cl, rqstr, rqstrlen);
    flatten_qstring(cl,  buf, NSTRING);
    flatten_qstring(nl, nbuf, NSTRING);
    flatten_qstring(pl, pbuf, NSTRING);

    /* 
     * Once upon a time, is_quote used to return the length of the quote
     * string that it had found. One day, not long ago, black hand came
     * and changed all that, and made is_quote return a quote string
     * divided in several fields, making the algorithm much more
     * complicated. Fortunately black hand left a few comments in the
     * source code to make it more understandable. Because of this change
     * we need to compute the lengths of the quote strings separately
     */
    c =  buf &&  buf[0] ? strlen(buf)  : 0;
    n = nbuf && nbuf[0] ? strlen(nbuf) : 0;
    p = pbuf && pbuf[0] ? strlen(pbuf) : 0;

    /*
     * When quote strings contain only blank spaces (ascii code 32) the
     * above count is equal to the length of the quote string, but if
     * there are TABS, the length of the quote string as seen by the user
     * is different than the number that was just computed.  Because of
     * this we demand a recount (hmm.. unless you are in Florida, where
     * recounts are forbidden)
     */

    NewP = strlenis(pbuf);
    NewC = strlenis(buf);
    NewN = strlenis(nbuf);

    /* 
     * For paragraphs with spaces in the first line, but no space in the
     * quote string of the second line, we make sure we choose the quote
     * string without a space at the end of it.
     */
    if ((NLine && !NLine[0])
	&& ((PLine && !PLine[0]) 
	     || (((same = same_qstring(pl, cl)) != 0)
			&& (same != count_levels_qstring(cl)))))
	cl = qs_remove_trailing_spaces(cl);
    else 
      if (NewC > NewN){
	int agree = 0;
        for (j = 0; (j < n) && (GLine[j] == NLine[j]); j++);
	clength = j;
	/* clength is the common length in which Gline and Nline agree */
	/* j < n means that they do not agree fully */
	/* GLine = "   \tText"
	   NLine = "   Text" */
	if(j == n)
	   agree++;
	if (clength < n){ /* see if buf and nbuf are padded with spaces and tabs */
	   for (i = clength; i < n && ISspace(NLine[i]); i++);
	   if (i == n){/* padded NLine until the end of spaces? */
	      for (i = clength; i < c && ISspace(GLine[i]); i++);
		if (i == c) /* Padded CLine until the end of spaces? */
		   agree++;
	   }
	}
        if (agree){
           for (j = clength; j < c && ISspace(GLine[j]); j++);
           if (j == c){

	/* 
	 * If we get here, it means that the current line has the same
	 * quote string (visually) than the next line, but both of them
	 * are padded with different amount of TABS or spaces at the end.
	 * The current line (GLine) has more spaces/TABs than the next
	 * line. This is the typical situation that is found at the
	 * begining of a paragraph. We need to check this, however, by
	 * checking the previous line. This avoids that we confuse
	 * ourselves with being in the last line of a paragraph.
	 * Example when it should not free_qs(cl)
	 * "    Text in Paragraph 1" (PLine)
	 * "    Text in Paragraph 1" (GLine)
	 * "  Other Paragraph Number 2" (NLine)
	 *
	 * Example when it should free_qs(cl):
	 * ":) "		(PLine) p = 3, j = 3
	 * ":)   Text"		(GLine) c = 5
	 * ":) More text" 	(NLine) n = 3
	 *
	 * Example when it should free_qs(cl):
	 * ":) "		(PLine)	p =  3, j = 3
	 * ":) > > >   Text"	(GLine)	c = 11
	 * ":) > > > More text" (NLine)	n =  9
	 *
	 * Example when it should free_qs(cl):
	 * ":) :) "		(PLine)	p =  6, j = 3
	 * ":) > > >   Text"	(GLine)	c = 11
	 * ":) > > > More text" (NLine)	n =  9
	 *
	 * Example when it should free_qs(cl):
	 * ":) > > >     "	(PLine)	p = 13, j = 11
	 * ":) > > >   Text"	(GLine)	c = 11
	 * ":) > > > More text" (NLine)	n =  9
	 *
	 * The following example is very interesting. The "Other Text"
	 * line below should free the quote string an make it equal to the
	 * quote string of the line below it, but any algorithm trying
	 * to advance past that line should make it stop there, so
	 * we need one more check, to check the raw quote string and the
	 * processed quote string at the same time.
	 * FREE qs in this example.
	 * "   Some Text"	(PLine) p = 3, j = 0
	 * "\tOther Text"	(GLine) c = 1
	 * "   More Text"	(NLine) n = 3
	 *
	 */

	      for (j = 0; (j < p) && (GLine[j] == PLine[j]); j++);
	      if (((p == c) && ((j != p) && NLine[n]))
		 || ((p != c) && NLine[n])){
		if(!get_indent_raw_line(q, PLine, nbuf, NSTRING, p, plb)
		  || NewP + strlenis(nbuf) != NewC){
		  free_qs(&cl);
		  cl = copy_qs(nl);
		}
	      }
           }   
        }
      }

    free_qs(&nl);
    free_qs(&pl);

    return cl;
}

/*
 * Given a line, an initial position, and a quote string, we advance the
 * current line past the quote string, including arbitraty spaces
 * contained in the line, except that it removes trailing spaces. We do
 * not handle TABs, if any, contained in the quote string. At least not
 * yet.
 *
 * Arguments: q - quote string
 *	      l - a line to process
 *	      i - position in the line to start processing. i = 0 is the
 *		  begining of that line.
 */
int
advance_quote_string(q, l, i)
    char *q;
    char  l[NSTRING];
    int   i;
{
    int n = 0, j = 0, is = 0, es = 0;
    int k, m, p, adv;
    char qs[NSTRING] = {'\0'};

    if(!q || !*q)
      return(0);

    for (p = strlen(q); (p > 0) && (q[p - 1] == ' '); p--, es++);
    if (!p){  /* string contains only spaces */
	for (k = 0; l[i + k] == ' '; k++);
	k -= k % es;
	return k;
    }
    for (is = 0; q[is] == ' '; is++); /* count initial spaces */
    for (m = 0 ; is + m < p ; m++)
	qs[m] = q[is + m];   /* qs = quote string without any space at the end */
		      /* advance as many spaces as there are at the begining */
    for (k = 0; l[i + j] == ' '; k++, j++);
			/* now find the visible string in the line */
    for (m = 0; qs[m] && l[i + j] == qs[m]; m++, j++);
    if (!qs[m]){	/* no match */
	/* 
	 * So far we have advanced at least "is" spaces, plus the visible
	 * string "qs". Now we need to advance the trailing number of
	 * spaces "es". If we can do that, we have found the quote string.
	 */
      for (p = 0; l[i + j + p] == ' '; p++);
      adv = advance_quote_string(q, l, i + j + ((p < es) ? p : es));
      n = ((p < es) ? 0 : es) + k + m + adv;
    }
    return n;
}

/* 
 * This function returns the effective length in screen of the quote
 * string. If the string contains a TAB character, it is added here, if
 * not, the length returned is the length of the string
 */

int
strlenis(qstr)
char *qstr;
{
  int i, rv = 0;

  for (i = 0; qstr && qstr[i]; i++)
      rv += ((qstr[i] == TAB) ? (~rv & 0x07) + 1 : 1);

  return rv;
}

/* Justify the entire buffer instead of just a paragraph */
fillbuf(f, n)
int f, n;
{
    int i, lastflagsave;
    LINE *eobline;
    REGION region;

    if(curbp->b_mode&MDVIEW){		/* don't allow this command if	*/
	return(rdonly());		/* we are in read only mode	*/
    }
    else if (fillcol == 0) {		/* no fill column set */
	mlwrite("No fill column set", NULL);
	return(FALSE);
    }

    if((lastflag & CFFILL) && (lastflag & CFFLBF)){
	/* no use doing a full justify twice */
	thisflag |= (CFFLBF | CFFILL);
	return(TRUE);
    }

    /* record the pointer of the last line */
    if(gotoeob(FALSE, 1) == FALSE)
      return(FALSE);

    eobline = curwp->w_dotp;		/* last line of buffer */
    if(!llength(eobline))
      eobline = lback(eobline);

    /* and back to the beginning of the buffer */
    gotobob(FALSE, 1);

    thisflag |= CFFLBF; /* CFFILL also gets set in fillpara */

    if(!Pmaster)
      sgarbk = TRUE;
    
    curwp->w_flag |= WFMODE;

    /*
     * clear the kill buffer, that's where we'll store undo
     * information, we can't do the fill buffer because
     * fillpara relies on its contents
     */
    kdelete();
    curwp->w_doto = 0;
    getregion(&region, eobline, llength(eobline));

    /* Put full message in the kill buffer for undo */
    if(!ldelete(region.r_size, kinsert))
      return(FALSE);

    /* before yank'ing, clear lastflag so we don't just unjustify */
    lastflag &= ~(CFFLBF | CFFILL);

    /* Now in kill buffer, bring back text to use in fillpara */
    yank(FALSE, 1);

    gotobob(FALSE, 1);

    /* call fillpara until we're at the end of the buffer */
    while(curwp->w_dotp != curbp->b_linep)
      if(!(fillpara(FALSE, 1)))
	return(FALSE);
    
    return(TRUE);
}


fillpara(f, n)	/* Fill the current paragraph according to the current
		   fill column						*/

int f, n;	/* deFault flag and Numeric argument */

{
    int	    i = 0, j, c, qlen, word[NSTRING], same_word, qlenis,
	    spaces, word_len, line_len, line_last, qn, indlen, qi, pqi;
    char   *qstr, qstr2[NSTRING], tbuf[NSTRING], ind_str[NSTRING], 
	   *qstrfl, qstrfl2[NSTRING], quoid[NSTRING];
    LINE   *eopline;
    REGION  region;
    QSTRING_S *tl;

    if(curbp->b_mode&MDVIEW){		/* don't allow this command if	*/
	return(rdonly());		/* we are in read only mode	*/
    }
    else if (fillcol == 0) {		/* no fill column set */
	mlwrite("No fill column set", NULL);
	return(FALSE);
    }
    else if(curwp->w_dotp == curbp->b_linep) /* don't wrap! */
      return(FALSE);

    /* record the pointer to the line just past the EOP */
    if(gotoeop(FALSE, 1) == FALSE)
      return(FALSE);

    eopline = curwp->w_dotp;		/* first line of para */
    /* and back to the beginning of the paragraph */
    gotobop(FALSE, 1);
    setimark(FALSE, 1);  /* Remember this spot in case we unjustify */

    /*
     * When a paragraph has special indentation, we will get two quote
     * strings. One from the first line of the paragraph, and one from
     * the last line of the paragraph. We will need to use both when
     * we justify.
     *
     * Here's a model of what we will code:
     *
     *      +-------+-------+-+-----+
     *      | qstrfl|ind_str|X| text|  
     *      +-----+-+-------+-+-----+
     *      | qstr| *(space)|X| text|
     *      +-----+---------+-+-----+
     *
     * Here X represents 1 space if it exists after ind_str and
     * "*(space)" represent a variable amount of space that is put there
     * to pad text so that it will align correctly when justified.
     */
    indlen = indent_match(default_qstr(), curwp->w_dotp, ind_str, NSTRING, 0);
    qstrfl = (quote_match(default_qstr(), curwp->w_dotp, qstrfl2, NSTRING, 0)
            && *qstrfl2) ? qstrfl2 : NULL;
    if (qstrfl)
       for (; (i < NSTRING) && (quoid[i] = qstrfl[i]); i++);
    if (indlen)
       for (j = 0; ((i + j) < NSTRING) && (quoid[i] = ind_str[j]); i++,j++); 
    quoid[i] = '\0';
    qi = quoid && quoid[0] ? strlen(quoid) : 0;
    if (indlen)			 /* strip trailing spaces */
       for (;ISspace(quoid[qi - 1]); qi--); 
    quoid[qi] = '\0';     /* we have closed quoid at "X" in the first line */
     
    if (strlenis(quoid) > fillcol)
       return FALSE;          /* Too wide, we can't justify this! */
      
     /* determine if we're justifying quoted text or not */
    qstr =  quote_match(default_qstr(), curwp->w_dotp, qstr2, NSTRING, 0)
  	    && *qstr2 ? qstr2 : NULL;
    /* In some situations, like in the following paragraph, qstr can be non
     * empty as returned above,  when indeed it is empty. Fix it!.

        * Item #1
        * Item #2
        continuation of item #2
     */
    if (qstr && indlen){
	for (i = strlen(qstr) - 1; ISspace(qstr[i]); i--);
	qstr[i + 1] = '\0';
    }

    qlen   = qstr ? strlen(qstr)   : 0;
    qlenis = qstr ? strlenis(qstr) : 0;
  
    /*
     * Compare effective lengths of quoid and qstr to decide how much space
     * we need to use to pad with.
     */
    if (indlen && ((j = strlenis(quoid) - strlenis(qstr)) > 0)){
       pqi = qstr ? strlen(qstr) : 0;
       for (i = 0; (i < j) && (qstr2[pqi + i] = ' '); i++);
       if (ISspace(ind_str[indlen - 1]))
          qstr2[pqi + i++] = ' ';
       qstr2[pqi + i] = '\0';
       if (!qstr)
          qstr = qstr2;
    }

    /* let yank() know that it may be restoring a paragraph */
    thisflag |= CFFILL;

    if(!Pmaster)
      sgarbk = TRUE;
    
    curwp->w_flag |= WFMODE;

    /* cut the paragraph into our fill buffer */
    fdelete();
    curwp->w_doto = 0;
    getregion(&region, eopline, llength(eopline));
    if(!ldelete(region.r_size, finsert))
      return(FALSE);

    /* Now insert it back wrapped */
    spaces = word_len = line_len = same_word = i = 0;

    /* Beginning with leading quoting... */
    if(qstrfl){
	while((tbuf[line_len] = qstrfl[line_len]) == fremove(line_len))
	  linsert(1, qstrfl[line_len++]);
	/* 
	 * The only way that at the end of the above loop we don't have
	 * line_len == qlen is that there are trailing spaces or TABS
	 * which could not be accounted in the qstr in is_quote or other
	 * functions before we got here. Now we enter the common part of
	 * the quote string in the first line and the rest is only spaces
	 * (or TABS) that need to be entered, which are left to the loop
	 * following this "if" statement
	 */
	i = line_len;				/* start next loop from here */
	tbuf[line_len] = '\0';			/* closing tbuf...	*/
	line_len = strlenis(tbuf);		/* we demand a recount! */
	line_last = ' ';			/* no word-flush space! */
    }

    /* ...followed by the indent string, if any */
    if (indlen){
	for (i, j = 0; (c = fremove(i)) && ind_str[j]; i++, j++){
	    linsert(1, line_last = c);
	    line_len += ((c == TAB) ? (~line_len & 0x07) + 1 : 1);
        }
	line_last = ' ';			/* no word-flush space! */
    }

    /* ...and leading white space */
    for(i; (c = fremove(i)) == ' ' || c == TAB; i++){
	linsert(1, line_last = c);
	line_len += ((c == TAB) ? (~line_len & 0x07) + 1 : 1);
    }

    /* then digest the rest... */
    while((c = fremove(i++)) > 0){
	switch(c){
	  case '\n' :
	    i += qlen;				/* skip next quote string */
	    if(!spaces)
	      spaces++;
	    same_word = 0;
	    break;

	  case TAB :
	  case ' ' :
	    spaces++;
	    same_word = 0;
	    break;

	  default :
	    if(spaces){				/* flush word? */
		if((line_len - qlenis > 0)
		   && line_len + word_len + 1 > fillcol
		   && ((ISspace(line_last))
		       || (linsert(1, ' ')))
		   && (line_len = fpnewline(qstr)))
		  line_last = ' ';	/* no word-flush space! */

		if(word_len){			/* word to write? */
		    if(line_len && !ISspace(line_last)){
			linsert(1, ' ');	/* need padding? */
			line_len++;
		    }

		    line_len += word_len;
		    for(j = 0; j < word_len; j++)
		      linsert(1, line_last = word[j]);

		    if(spaces > 1 && strchr(".?!:;\")", line_last)){
			linsert(2, line_last = ' ');
			line_len += 2;
		    }

		    word_len = 0;
		}

		spaces = 0;
	    }

	    if(word_len + 1 >= NSTRING){
		/* Magic!  Fake that we output a wrapped word */
		if((line_len - qlenis > 0) && !same_word++){
		    if(!ISspace(line_last))
		      linsert(1, ' ');
		    line_len = fpnewline(qstr);
		}

		line_len += word_len;
		for(j = 0; j < word_len; j++)
		  linsert(1, word[j]);

		word_len  = 0;
		line_last = ' ';
	    }

	    word[word_len++] = c;
	    break;
	}
    }

    if(word_len){
	if((line_len - qlenis > 0) && (line_len + word_len + 1 > fillcol)){
	    if(!ISspace(line_last))
	      linsert(1, ' ');
          if (line_len && (line_len != qlenis))
             (void) fpnewline(qstr);
	}
	else if(line_len && !ISspace(line_last))
	  linsert(1, ' ');

	for(j = 0; j < word_len; j++)
	  linsert(1, word[j]);
    }

    /* Leave cursor on first char of first line after paragraph */
    curwp->w_dotp = lforw(curwp->w_dotp);
    curwp->w_doto = 0;

    return(TRUE);
}


/*
 * fpnewline - output a fill paragraph newline mindful of quote string
 */
fpnewline(quote)
char *quote;
{
    int i;

    lnewline();
    for(i = 0; quote && quote[i]; i++)
      linsert(1, quote[i]);

    return strlenis(quote);
}

int
is_indent (word, plb)
  char word[NSTRING];
  int  plb;
{
  int i = 0, finished = 0, c, nxt, j, k, digit = 0, bdigits = -1, alpha = 0;

   if (!word || !word[0])
      return i;

   for (i = 0, j = 0; ISspace(word[i]); i++, j++);
   while ((i < NSTRING - 2) && !finished){
	switch (c = now(word,i)){
	    case NBSP:
	    case TAB :
	    case ' ' : for (; ISspace(word[i]); i++);
		      if (!is_indent_char(now(word,i)))
			finished++;
		   break;

	    case '+' :
	    case '.' :
	    case ']' :
	    case '*' :
	    case '}' :
	    case '-' :
	    case RPAREN:
		   nxt = next(word,i);
	           if (((c == '.') && allowed_after_period(nxt) && alpha)
			|| ((c == '*') && allowed_after_star(nxt))
		        || ((c == '}') && allowed_after_braces(nxt))
		        || ((c == '-') && allowed_after_dash(nxt))
		        || ((c == '+') && allowed_after_dash(nxt))
		        || ((c == RPAREN) && allowed_after_parenth(nxt))
		        || ((c == ']') && allowed_after_parenth(nxt)))
		      i++;
		   else
		      finished++;
		   break;

	    default : if (is_a_digit(c) && plb){
			if (bdigits < 0)
			   bdigits = i;  /* first digit */
			for (k = i; is_a_digit(now(word,k)); k++);
			if (k - bdigits > 2){ /* more than 2 digits? */
			   i = bdigits;	/* too many! */
			   finished++;
			}
			else{ 
			   if(allowed_after_digit(now(word,k),word,k)){
			     alpha++;
			     i = k;
			   }
			   else{
			     i = bdigits;
			     finished++;
			   }
			}
		      }
		      else
			finished++;
		   break;
	}
   }
   if (i == j)
      i = 0;  /* there must be something more than spaces in an indent string */
   return i;   
}

/*
 * If there is an indent string this function returns 
 * its length
 */
int
indent_match(q, l, buf, buflen, raw)
    char *q;
    LINE *l;
    char *buf;
    int   buflen;
    int   raw;
{
    char GLine[NSTRING] = {'\0'};
    int  i, k, plb;

    k = quote_match(q,l, buf, buflen, raw);

    linencpy(GLine, l, NSTRING);

    plb = (lback(l) != curbp->b_linep) ? lisblank(lback(l)) : 1;
    if (!plb){
	i = llength(lback(l)) - 1;
        for (; i >= 0 && ISspace(lgetc(lback(l), i).c); i--);
	if (EOLchar(lgetc(lback(l), i).c))
	   plb++;
    }

    return get_indent_raw_line(q, GLine, buf, buflen, k, plb);
}

int
get_indent_raw_line(q, GLine, buf, buflen, k, plb)
    char *q;
    char *GLine;
    char *buf;
    int   buflen;
    int   k;
    int   plb;
{
    int i, j;

    i = is_indent(GLine+k, plb);

    for (j = 0; (j < i) && (j < buflen) && (buf[j] = GLine[j + k]); j++);
    buf[j] = '\0';

    return i;
}

deletepara(f, n)	/* Delete the current paragraph */

int f, n;	/* deFault flag and Numeric argument */

{
    if(curbp->b_mode&MDVIEW){           /* don't allow this command if  */
        return(rdonly());               /* we are in read only mode     */
    }

    if(!lisblank(curwp->w_dotp))
      gotobop(FALSE, 1);

    curwp->w_markp = curwp->w_dotp;
    curwp->w_marko = curwp->w_doto;

    gotoeop(FALSE, 1);
    if (curwp->w_dotp != curbp->b_linep){ /* if we are not at the end of buffer */
        curwp->w_dotp = lforw(curwp->w_dotp); /* get one more line */
        curwp->w_doto = 0; /* but only the beginning */
    }
    killregion(f,n);
          
    return(TRUE);
}

