/* This module was written by
 *
 * Eduardo Chappa (chappa@math.washington.edu)
 * http://www.math.washington.edu/~chappa/pine/
 *
 *  Original Version: November 1999
 *  Last Modified   : December 14, 2004
 *
 * Send bug reports about this module to the address above.
 */

#include "rules.h"

void free_token_value(token)
  TOKEN_VALUE **token;
{
   if(token && *token){
     if ((*token)->testxt)
	fs_give((void **)&((*token)->testxt));
     if((*token)->next)
	free_token_value(&((*token)->next));
     fs_give((void **)token);
   }
}

void free_condition(condition)
  CONDITION_S **condition;
{
   if(condition && *condition){
     if ((*condition)->tname)
	fs_give((void **)&((*condition)->tname));
     if ((*condition)->value)
	free_token_value(&((*condition)->value));
     if((*condition)->next)
	free_condition(&((*condition)->next));
     fs_give((void **)condition);
   }
}

void free_ruleaction(raction)
  RULEACTION_S **raction;
{
   if(raction && *raction){
     if ((*raction)->token)
	fs_give((void **)&((*raction)->token));
     if ((*raction)->function)
	fs_give((void **)&((*raction)->function));
     if ((*raction)->value)
	fs_give((void **)&((*raction)->value));
     fs_give((void **)raction);
   }
}

void free_rule(rule)
  RULE_S **rule;
{
   if(rule && *rule){
     free_condition(&((*rule)->condition));
     free_ruleaction(&((*rule)->action));
     fs_give((void **)rule);
   }
}

void free_rule_list(rule)
  RULELIST **rule;
{
  if(!*rule)
    return;

  if((*rule)->next)
    free_rule_list(&((*rule)->next));

  if((*rule)->prule)
    free_rule(&((*rule)->prule));

  fs_give((void **)rule);
}

void
free_parsed_rule_list(rule)
  PRULELIST_S **rule;
{
  if(!*rule)
    return;

  if((*rule)->next)
    free_parsed_rule_list(&((*rule)->next));

  if((*rule)->rlist)
    free_rule_list(&((*rule)->rlist));

  fs_give((void **)rule);
}

void *
alloc_mem (amount)
  size_t amount;
{
   void *genmem;
   memset(genmem = fs_get(amount), 0, amount);
   return genmem;
}

int
isolate_condition (data, cvalue, len)
  char *data, **cvalue;
  int  *len;
{
  char *p = data;
  int done = 0, error = 0, next_condition = 0, l;

  *cvalue = NULL;
  while (*p && !done){
	switch (*p){
	   case '_': *cvalue = advance_to_char(p,'}', STRICT, NULL);
		     if(*cvalue){
			strcat(*cvalue,"}");
			p += strlen(*cvalue);
		     }
		     else
			error++;
		     done++;
	   case ' ': p++;
		     break;
	   case '&': if (*(p+1) == '&'){	/* looking for && */
			p += 2;
			next_condition++;
		     }
		     else{
			error++;
			done++;
		     }
		     break;
	   case '=': /* looking for => or -> */
	   case '-': if ((*(p+1) == '>') && (!next_condition)){
		        is_save = (*p == '-');
			p += 2;
		     }
		     else
			error++;
		     done++;
		     break;
	   default : done++;
		     error++;
		     break;
	}
  }
  *len = p - data;
  return error ? -1 : (*cvalue ? 1 : 0);
}

TOKEN_VALUE *
parse_group_data (data, error)
  char *data;
  int *error;
{
  TOKEN_VALUE *rvalue;
  char *p;
  int offset, err = 0;

  if(error)
    *error = 0;

  if (!data)
     return (TOKEN_VALUE *) NULL;

  rvalue = (TOKEN_VALUE *) alloc_mem(sizeof(TOKEN_VALUE));
  if (p = advance_to_char(data,';', STRICT, &offset)){
      rvalue->testxt = cpystr(p);
      data += strlen(p) + 1 + offset;
      rvalue->next   = parse_group_data(data, error);
  }
  else if (p = advance_to_char(data,'}', STRICT, NULL))
      rvalue->testxt = cpystr(p);
  else{
      err++;
      free_token_value(&rvalue);
  }
  if (error)
    *error += err;
  return(rvalue);
}

CONDITION_S *
fill_condition(data)
  char *data;
{
  CONDITION_S *condition;
  int i, done, error = 0;
  char *group;

  for (i = 0, done = 0; !done && (i < NTOKENS); i++)
      done = strncmp(data,token_rules[i], strlen(token_rules[i])) ? 0 : 1;
  if (done){
     condition = (CONDITION_S *) alloc_mem(sizeof(CONDITION_S));
     condition->tname = cpystr(token_rules[--i]);
  }
  else
     return (CONDITION_S *)NULL;

  data += strlen(token_rules[i]);
  for (; *data && *data == ' '; data++);
  if (*data){
     for (i = 0, done = 0; !done && (i < NREL); i++)
       done = strncmp(data, rel_rules_test[i].value, 2) ? 0 : 1;
     if (done)
       condition->ttype = rel_rules_test[--i].ttype;
     else{
	 free_condition(&condition);
	 return (CONDITION_S *) NULL;
     }
  }
  data += 2;
  for (; *data && *data == ' '; data++);
  if (*data++ != '{'){
     free_condition(&condition);
     return (CONDITION_S *) NULL;
  }
  group = advance_to_char(data,'}', STRICT, &error); 
  if (group || (!group &&  error < 0)){
     condition->value = parse_group_data(data, &error);
     if(group && error)
	free_condition(&condition);
     if(group)
        fs_give((void **) &group);
  }
  else
     free_condition(&condition);
  return condition;
}

CONDITION_S *
parse_condition (data, eoc)
  char *data;
  int  *eoc;	/* end of condition, equal to -1 on error */
{
  CONDITION_S *condition = NULL;
  char *p = data, *cvalue;
  int len, error = 0, rv;

  if((rv = isolate_condition(data, &cvalue, &len)) > 0){
    if(condition = fill_condition(cvalue))
       condition->next = parse_condition(data+len, eoc);
    else
       error++;
  }
  *eoc += len;
  if (error)
    *eoc = -1;
  return condition;
}

RULEACTION_S *
parse_action (data, context)
  char *data;
  int context;
{
  int i, done;
  RULEACTION_S *raction = NULL;
  char *function, *p = data;

  if (!p)
     return (RULEACTION_S *) NULL;

  for (; *p && *p == ' '; p++);
  if (!*p)
     return (RULEACTION_S *) NULL;

  if (is_save){
     raction = (RULEACTION_S *) alloc_mem(sizeof(RULEACTION_S));
     raction->function = cpystr("_SAVE_");
     raction->value    = (TOKEN_VALUE *) alloc_mem(sizeof(TOKEN_VALUE));
     raction->context |= FOR_SAVE;
     raction->exec     = extended_value;
     raction->value->testxt = cpystr(p);
     return raction;
  }
  for (i = 0, done = 0; !done && (i < NFCN); i++)
       done = (strstr(p,rule_fcns[i].name) == p);
  p += done ? strlen(rule_fcns[--i].name) + 1 : 0;
  if(!*p || (rule_fcns[i].what_for && !(rule_fcns[i].what_for & context)))
     return (RULEACTION_S *) NULL;
  if (done){
     raction = (RULEACTION_S *) alloc_mem(sizeof(RULEACTION_S));
     raction->function = cpystr(rule_fcns[i].name);
     raction->context  = rule_fcns[i].what_for;
     raction->exec     = rule_fcns[i].execute;
     raction->value    = (TOKEN_VALUE *) alloc_mem(sizeof(TOKEN_VALUE));
     raction->value->testxt = advance_to_char(p,'}', STRICT, NULL);
     if(!raction->value->testxt)
       free_ruleaction(&raction);
     return raction;
  }

  done = (((function = strstr(p, "_TRIM_")) != NULL)
	  ? 1 : ((function = strstr(p, "_REXTRIM_")) != NULL)
	  ? 2 : ((function = strstr(p, "_REPLACE_")) != NULL)
	  ? 3 : 0);

  if(!function)
     return (RULEACTION_S *) NULL;

  *function = '\0';
   raction = (RULEACTION_S *) alloc_mem(sizeof(RULEACTION_S));
   raction->token = get_name_token(p);
  *function = '_';
   p += strlen(raction->token) + 1;
   for (; *p && *p == ' '; p++);
   if (!strncmp(p,":=",2))
      p += 2;
   else{
      free_ruleaction(&raction);
      return (RULEACTION_S *) NULL;
   }
   for (; *p && *p == ' '; p++);
   if (p != function){
      free_ruleaction(&raction);
      return (RULEACTION_S *) NULL;
   }
   p += done == 1 ? 6: 9;
   if (*p != '{'){
      free_ruleaction(&raction);
      return (RULEACTION_S *) NULL;
   }
   *p = '\0';
   for(i = 0; i < NFCN && strcmp(function, rule_fcns[i].name);i++);
   raction->function   = cpystr(function);
   raction->is_trim    = strcmp(function,"_TRIM_")    ? 0 : 1;
   raction->is_rextrim = strcmp(function,"_REXTRIM_") ? 0 : 1;
   raction->is_replace = strcmp(function,"_REPLACE_") ? 0 : 1;
   raction->context  = rule_fcns[i].what_for;
   raction->exec     = rule_fcns[i].execute;
   *p++ = '{';
   raction->value = parse_group_data(p, NULL);
   if(!raction->value->testxt)
      free_ruleaction(&raction);
   return raction;
}

RULE_S *
parse_rule (data, context)
  char *data;
  int context;
{
  RULE_S *prule;	/*parsed rule */
  int len = 0;
  
  if (!(prule = (RULE_S *) alloc_mem(sizeof(RULE_S))) ||
	!(prule->condition = parse_condition(data, &len)) ||
	!(prule->action = parse_action(data+len, context)))
      free_rule(&prule);

  return prule;
}

RULELIST *
get_rule_list(list, context, i)
  char **list;
  int context, i;
{
  RULE_S *rule;
  RULELIST *trulelist = NULL;

  if (list[i] && *list[i]){
     if(rule = parse_rule(list[i], context)){
	trulelist  = (RULELIST *)alloc_mem(sizeof(RULELIST));
	trulelist->prule = rule;
	trulelist->next = get_rule_list(list, context, i+1);
     }
     else
	trulelist = get_rule_list(list, context, i+1);
  }
  return trulelist;
}

PRULELIST_S *
add_prule(rule_list, rule)
  PRULELIST_S *rule_list;
  PRULELIST_S *rule;
{
   if (!rule_list)
      rule_list = (PRULELIST_S *) alloc_mem(sizeof(PRULELIST_S));

   if(rule_list->next)
     rule_list->next = add_prule(rule_list->next, rule);
   else{
     if (rule_list->rlist)
	rule_list->next = rule;
     else
	rule_list = rule;
   }
   return rule_list;
}  

void
add_rule(code, context)
  int code, context;
{
  char **list = ps_global->vars[code].current_val.l;
  PRULELIST_S *prulelist, *trulelist, *orulelist;

  if (list && *list && **list){
     trulelist = (PRULELIST_S *)alloc_mem(sizeof(PRULELIST_S));
     trulelist->varnum = code;
     if (trulelist->rlist = get_rule_list(list, context, 0))
        ps_global->rule_list = add_prule(ps_global->rule_list, trulelist);
     else
	free_parsed_rule_list(&trulelist);
  }
}

void
create_rule_list(void)
{
  add_rule(V_THREAD_DISP_STYLE_RULES,FOR_RULE|FOR_THREAD);
  add_rule(V_THREAD_INDEX_STYLE_RULES,FOR_RULE|FOR_THREAD);
  add_rule(V_COMPOSE_RULES,FOR_RULE|FOR_COMPOSE);
  add_rule(V_FORWARD_RULES,FOR_RULE|FOR_COMPOSE);
  add_rule(V_INDEX_RULES,FOR_RULE|FOR_INDEX);
  add_rule(V_REPLACE_RULES,FOR_RULE|FOR_REPLACE);
  add_rule(V_REPLY_INDENT_RULES,FOR_RULE|FOR_COMPOSE);
  add_rule(V_REPLY_LEADIN_RULES,FOR_RULE|FOR_REPLY_INTRO);
  add_rule(V_RESUB_RULES,FOR_RULE|FOR_RESUB|FOR_TRIM);
  add_rule(V_SAVE_RULES,FOR_RULE|FOR_SAVE);
  add_rule(V_SMTP_RULES,FOR_RULE|FOR_COMPOSE);
  add_rule(V_SORT_RULES,FOR_RULE|FOR_SORT);
  add_rule(V_STARTUP_RULES,FOR_RULE|FOR_STARTUP);
}

int
condition_contains_token(condition, token)
  CONDITION_S *condition;
  char *token;
{
  if (!condition)
    return 0;

  if (!strcmp(condition->tname, token))
     return 1;
  else
     return condition_contains_token(condition->next, token);
}

RULELIST *
get_rulelist_from_code(code, list)
  int code;
  PRULELIST_S *list;
{
  if (!list)
     return (RULELIST *) NULL;

  if(list->varnum == code)
     return list->rlist;
  else
     return get_rulelist_from_code(code, list->next);
}   

char *
test_rule(rlist, ctxt, env, n)
  RULELIST *rlist;
  int ctxt, *n;
  ENVELOPE *env;
{
  char *result;

  if(!rlist)
     return NULL;

  if (result = process_rule(rlist->prule, ctxt, env))
      return result;
  else{
       (*n)++;
       return test_rule(rlist->next, ctxt, env, n);
  } 
}

RULE_S *
get_rule (rule, n)
  RULELIST *rule;
  int n;
{
  if (!rule)
     return (RULE_S *) NULL;

  return n ? get_rule(rule->next, n-1) : rule->prule;
}

/* get_result_rule:
 * Parameters: list: the list of rules to be passed to the function to check
 *             rule_context: context of the rule
 *             env : envelope used to check the rule, if needed.
 *
 * Returns: The value of the first rule that is satisfied in the list, or
 *          NULL if not. This function should be called in the following 
 *          way (notice that memory is freed by caller).
 *
 * You should use this function to obtain the result of a rule. You can
 * also call directly "process_rule", but I advice to use this function if
 * there's no difference on which function to call.

   RULE_RESULT *rule;

   rule = (RULE_RESULT *) 
           get_result_rule(V_SOME_RULE, context, envelope);

   if (rule){ 
       assign the value of rule->result;
       if (rule->result)
          fs_give((void **)&rule->result);
       fs_give((void **)&rule);
   }
 */

RULE_RESULT *
get_result_rule(code, rule_context, env)
 int code, rule_context;
 ENVELOPE *env;
{
    char  *rule_result;
    RULE_RESULT *rule = NULL;
    RULELIST *rlist;
    int n = 0;

    rlist = get_rulelist_from_code(code, ps_global->rule_list);
    if (rlist){
       rule_result = test_rule(rlist, rule_context, env, &n);
       if (rule_result && *rule_result){
          rule = (RULE_RESULT *) fs_get (sizeof(RULE_RESULT));
          rule->result = rule_result;
          rule->number = n;
       }
    }
    return rule;
}

/* process_rule:
   Parameters: rule_data, is a rule. It's obtained as
                rule_data = ps_global->VAR_SOME_RULE[n], for
                some integer n
                rule_context: context of the rule, and
                env: An envelope if needed.

   Returns   : The value of the processed rule_data if the processing was 
               successful and matches context and possibly the envelope, or
               NULL if there's no match
 */

char *
process_rule (prule, rule_context, env)
 RULE_S *prule;
 int  rule_context;
 ENVELOPE *env;
{
    int rv;
    char *result = NULL;
    CONDITION_S *condition;

   if(!prule)
     return NULL;

   for(condition = prule->condition; 
		   condition && 
		   (rv = test_condition(condition, rule_context, env));
		   condition = condition->next);

   if(rv && !condition)
      result = (prule->action->exec)(prule->action, rule_context, env);

   return result;
}

TOKEN_VALUE *
copy_parsed_value(value, ctxt, env)
   TOKEN_VALUE *value;
   int ctxt;
   ENVELOPE *env;
{
   TOKEN_VALUE *tval = NULL;

   if(!value)
      return NULL;

   if(value->testxt){
     tval = (TOKEN_VALUE *) alloc_mem(sizeof(TOKEN_VALUE));
     tval->testxt = detoken_src(value->testxt, ctxt, env, NULL, NULL, NULL);
   }
   if(value->next)
     tval->next = copy_parsed_value(value->next, ctxt, env);

   return tval;
}

int
test_condition(condition, rule_context, env)
CONDITION_S *condition;
int rule_context;
ENVELOPE *env;
{
   int next_step;
   TOKEN_VALUE *group;

   group = copy_parsed_value(condition->value, rule_context, env);
   next_step = (*rel_rules_test[condition->ttype].execute)(condition, group, env, rule_context);
   free_token_value(&group);
   return next_step;
}

/* returns the name of the token it found or NULL if there is no token, the
 * real value of the token is obtained by calling the detoken_src function.
 */ 

char *
get_name_token (condition)
char *condition;
{
  char *p = NULL, *q, *s;

    if ((q = strchr(condition,'_')) && (s = strchr(q+1,'_'))){
	char c = *++s;
	*s = '\0';
	 p = cpystr(q);
	*s = c;
    }
    return p;
}

/* This function tests if a string contained in the variable "group" is
 * in the "condition"
 */
int test_in (condition, group, env, context)
TOKEN_VALUE *group;
CONDITION_S *condition;
ENVELOPE *env;
int context;
{
 int rv = 0;
 char *test;
 TOKEN_VALUE *test_group = group;

 test = detoken_src(condition->tname, context, env, NULL, NULL, NULL);
 if (test){
    while (rv == 0 && test_group){
       if(!*test || strstr(test_group->testxt, test))
	  rv++;
       else
	  test_group = test_group->next;
    }
 }
 return rv;
}

int test_ni (condition, group, env, context)
TOKEN_VALUE *group;
CONDITION_S *condition;
ENVELOPE *env;
int context;
{
 int rv = 0;
 char *test;
 TOKEN_VALUE *test_group = group;

 test = detoken_src(condition->tname, context, env, NULL, NULL, NULL);
 if (test){
    if(!test_group)
      rv++;
    while (rv == 0 && test_group){
       if(!*test_group->testxt || strstr(test, test_group->testxt))
	  rv++;
       else
	  test_group = test_group->next;
    }
 }
 return rv;
}

int test_not_in (condition, group, env, context)
TOKEN_VALUE *group;
CONDITION_S *condition;
ENVELOPE *env;
int context;
{
 return !test_in(condition, group, env, context);
}

int test_not_ni (condition, group, env, context)
TOKEN_VALUE *group;
CONDITION_S *condition;
ENVELOPE *env;
int context;
{
 return !test_ni(condition, group, env, context);
}

int test_eq (condition, group, env, context)
CONDITION_S *condition;
TOKEN_VALUE *group;
ENVELOPE *env;
int context;
{
 int rv = 0;
 char *test;
 TOKEN_VALUE *test_group = group;

 test = detoken_src(condition->tname, context, env, NULL, NULL, NULL);
 if (test){
    while (rv == 0 && test_group){
       if((!*test && !*test_group->testxt) || !strcmp(test_group->testxt, test))
	  rv++;
       else
	  test_group = test_group->next;
    }
 }
 return rv;
}

int test_not_eq (condition, group, env, context)
CONDITION_S *condition;
TOKEN_VALUE *group;
ENVELOPE *env;
int context;
{
 return !test_eq(condition, group, env, context);
}

char *
do_trim (test, tval)
  char *test;
  TOKEN_VALUE *tval;
{
   char *begin_text;
   int offset = 0;

   if (!tval)
      return test;

   while(begin_text = strstr(test+offset,tval->testxt)){
      strcpy(begin_text, begin_text+strlen(tval->testxt));
      offset = begin_text - test;
   }

   return do_trim(test, tval->next);
}

char *
trim (action, context, env)
 RULEACTION_S *action;
 int context;
 ENVELOPE *env;
{
 char *begin_text, *test;
 RULEACTION_S *taction = action;
 int offset;

 if (taction->context & context){
    test = detoken_src(taction->token, context, env, NULL, NULL, NULL);
    if (test)
       test = do_trim(test, taction->value);
    return test;
 }
 return NULL;
}


char *
do_rextrim (test, tval)
  char *test;
  TOKEN_VALUE *tval;
{
   char *begin_text, *trim_text;
   int offset = 0;

   if (!tval)
      return test;

   trim_text = expand(test, tval->testxt);
   while(trim_text && (begin_text = strstr(test+offset,trim_text))){
      strcpy(begin_text, begin_text+strlen(trim_text));
      offset = begin_text - test;
   }

   return do_rextrim(test, tval->next);
}

char *
rextrim (action, context, env)
 RULEACTION_S *action;
 int context;
 ENVELOPE *env;
{
 char *begin_text, *trim_text, *test;
 RULEACTION_S *taction = action;
 TOKEN_VALUE **token = NULL;
 int offset;

 if (taction->context & context){
    test = detoken_src(taction->token, context, env, NULL, NULL, NULL);
    if (test){
	test = do_rextrim(test, taction->value);
    }
    return test;
  }
  else
     return NULL;
}

char *
raw_value (action, context,env)
RULEACTION_S *action;
int context;
ENVELOPE *env;
{
return (action->context & context) ? cpystr(action->value->testxt) : NULL;
}

char *
extended_value (action, ctxt,env)
RULEACTION_S *action;
int ctxt;
ENVELOPE *env;
{
return (action->context & ctxt) 
	? detoken_src(action->value->testxt, ctxt, env, NULL, NULL, NULL)
	: NULL;
}

/* advances given_string until it finds given_char  */
char *
advance_to_char(given_string,given_char, flag, error) 
 char *given_string;                          
 char given_char;
 int flag;
 int *error;
{
   char *b, *s, c;
   int i, err  = 0, quoted ;

   if (error) 
      *error = 0;

   if (!given_string || !*given_string)
       return NULL;

   b = s = cpystr(given_string);
   for(i = 0, quoted = 0, c = *s; c ; c = *++s){
      if(c == '\\'){
	 quoted++;
	 continue;
      }
      if(quoted){
	quoted = 0;
	if (c == given_char){
	  err += flag & STRICT ? 0 : 1;
	  err++;
	  break;
	}
	b[i++] = '\\';
      }
      if(c == given_char){
	 err += flag & STRICT ? 0 : 1;
	 break;
      }
      b[i++] = c;
   }
   b[i] = '\0';
   if (b && (strlen(b) == strlen(given_string)) && (flag & STRICT))
      return NULL;   /* character not found */

   if(b && !*b){
     fs_give((void **)&b);
     err = -1;
   }

   if (error)
      *error = err;

   return b;
}

/* Regular Expressions Support */

char *
expand (string, pattern)
char *string;
char *pattern;
{
 char tmp[1000];
 int error, i = 0;
 char *new_start, c;
 char *ret_string = NULL;
 regex_t preg;
 regmatch_t pmatch;
 
  if (error = regcomp(&preg, pattern, REG_EXTENDED)){
     regerror(error, &preg, tmp, 1000);
     return NULL;
  }

  if(((error = regexec(&preg, string, 1, &pmatch, 0)) != REG_NOMATCH)&& !error){
      c = string[pmatch.rm_eo];
      string[pmatch.rm_eo] = '\0';
      ret_string = cpystr(string+pmatch.rm_so);
      string[pmatch.rm_eo] = c;
  }
  return ret_string;
}
