#include "headers.h"
#ifndef _REGEX_H_
#include <regex.h>
#endif

int is_save;	/* this rule has the form condition -> folder */
typedef struct {
	char	*value;
	int	type;
} RULE_ACTION;


RULE_RESULT *get_result_rule PROTO ((int, int, ENVELOPE *));
char	*test_rule	PROTO ((RULELIST *, int, ENVELOPE *, int *));
char	*process_rule	PROTO ((RULE_S *, int, ENVELOPE *));
int	test_condition	PROTO ((CONDITION_S *, int, ENVELOPE *));
int	test_in		PROTO ((CONDITION_S *, TOKEN_VALUE *, ENVELOPE *, int));
int	test_ni		PROTO ((CONDITION_S *, TOKEN_VALUE *, ENVELOPE *, int));
int	test_not_in	PROTO ((CONDITION_S *, TOKEN_VALUE *, ENVELOPE *, int));
int	test_not_ni	PROTO ((CONDITION_S *, TOKEN_VALUE *, ENVELOPE *, int));
int	test_eq		PROTO ((CONDITION_S *, TOKEN_VALUE *, ENVELOPE *, int));
int	test_not_eq	PROTO ((CONDITION_S *, TOKEN_VALUE *, ENVELOPE *, int));
char	*trim		PROTO ((RULEACTION_S *, int, ENVELOPE *));
char	*rextrim	PROTO ((RULEACTION_S *, int, ENVELOPE *));
char	*raw_value	PROTO ((RULEACTION_S *, int, ENVELOPE *));
char	*extended_value	PROTO ((RULEACTION_S *, int, ENVELOPE *));
char	*expand		PROTO((char *, char *));
char	*get_name_token PROTO((char *));
char	*advance_to_char PROTO ((char *, char, int, int *));

/* Separators:
 *
 * A separator is a string that separates the rule condition with the rule
 * action. Below is the list of separators
 *
 */

#define  SAVE_TO_SEP  "->"
#define  APPLY_SEP    "=>"

/*------- Definitions of tokens -------*/
/*------ Keep the list alphabetically sorted, thanks -------*/

#define ADDR_TOKEN	"_ADDRESS_"
#define ADDCC_TOKEN	"_ADDRESSCC_"
#define ADDRECIP_TOKEN	"_ADDRESSRECIPS_"
#define ADDTO_TOKEN	"_ADDRESSTO_"
#define BCC_TOKEN	"_BCC_"
#define CC_TOKEN	"_CC_"
#define COLLECT_TOKEN	"_COLLECTION_"
#define FLAG_TOKEN	"_FLAG_"
#define FOLDER_TOKEN	"_FOLDER_"
#define FADDRESS_TOKEN	"_FORWARDADDRESS_"
#define FFROM_TOKEN	"_FORWARDFROM_"
#define FROM_TOKEN	"_FROM_"
#define LCC_TOKEN	"_LCC_"
#define NICK_TOKEN	"_NICK_"
#define ROLE_TOKEN	"_ROLE_"
#define SEND_TOKEN	"_SENDER_"
#define SUBJ_TOKEN	"_SUBJECT_"
#define THDDSPSTY_TOKEN	"_THREADSTYLE_"
#define THDNDXSTY_TOKEN	"_THREADINDEX_"
#define TO_TOKEN	"_TO_"

/* Mail related tokens 
 *
 * The following is an array with the list of tokens.
 */

char* token_rules[] = {
   FROM_TOKEN,
   NICK_TOKEN,
   ROLE_TOKEN,
   FOLDER_TOKEN,
   SUBJ_TOKEN,
   THDDSPSTY_TOKEN,
   THDNDXSTY_TOKEN,
   FLAG_TOKEN,
   COLLECT_TOKEN,
   THDDSPSTY_TOKEN,
   ADDR_TOKEN,
   TO_TOKEN,
   ADDTO_TOKEN,
   ADDCC_TOKEN,
   ADDRECIP_TOKEN,
   SEND_TOKEN,
   CC_TOKEN,
   LCC_TOKEN,
   BCC_TOKEN,
   FFROM_TOKEN,
   FADDRESS_TOKEN,
   NULL
};

#define NTOKENS  (sizeof(token_rules)/sizeof(token_rules[0]) - 1)

/*------ Definitions of relational operands -------------*/

typedef struct {
        char       *value;
	TestType    ttype;
        int        (*execute)();
} REL_TOKEN;

/* Relational Operands */
#define AND_REL     "&&"        /* For putting more than one condition  */
#define IN_REL      "<<"        /* For belonging relation */
#define NI_REL      ">>"        /* For contain relation   */
#define NOT_IN_REL  "!<"        /* Negation of IN_REL   */
#define NOT_NI_REL  "!>"        /* Negation of NI_REL   */
#define EQ_REL      "=="        /* Test of equality     */
#define NOT_EQ_REL  "!="        /* Test of inequality   */
#define OPEN_SET    "{"         /* Braces to open a set */
#define CLOSE_SET   "}"         /* Braces to close a set*/

REL_TOKEN rel_rules_test[] = {
   {EQ_REL,	Equal,		test_eq},
   {IN_REL,	Subset,		test_in},
   {NI_REL,	Includes,	test_ni},
   {NOT_EQ_REL, NotEqual,	test_not_eq},
   {NOT_IN_REL, NotSubset,	test_not_in},
   {NOT_NI_REL, NotIncludes,	test_not_ni},
   {NULL,	EndTypes,	NULL}
};

#define NREL  (sizeof(rel_rules_test)/sizeof(rel_rules_test[0]) - 1)

/*--- Context in which these variables can be used ---*/

typedef struct use_context {
    char        *name;
    int          what_for;
} USE_IN_CONTEXT;


static USE_IN_CONTEXT tokens_use[] = {
    {NICK_TOKEN,        FOR_SAVE},
    {FROM_TOKEN,        FOR_SAVE},
    {ROLE_TOKEN,        FOR_COMPOSE},
    {FOLDER_TOKEN,      FOR_SAVE|FOR_FOLDER|FOR_THREAD},
    {SUBJ_TOKEN,        FOR_SAVE|FOR_FOLDER},
    {FLAG_TOKEN,        FOR_SAVE|FOR_FLAG},
    {COLLECT_TOKEN,     FOR_SAVE|FOR_COMPOSE|FOR_FOLDER|FOR_THREAD},
    {THDDSPSTY_TOKEN,   FOR_THREAD},
    {THDNDXSTY_TOKEN,   FOR_THREAD},
    {ADDR_TOKEN,        FOR_SAVE|FOR_FOLDER},
    {TO_TOKEN,          FOR_SAVE},
    {ADDTO_TOKEN,       FOR_SAVE|FOR_COMPOSE},
    {ADDCC_TOKEN,       FOR_SAVE|FOR_COMPOSE},
    {ADDRECIP_TOKEN,    FOR_SAVE|FOR_COMPOSE},
    {SEND_TOKEN,        FOR_SAVE},
    {CC_TOKEN,          FOR_SAVE},
    {BCC_TOKEN,         FOR_COMPOSE},
    {LCC_TOKEN,         FOR_COMPOSE},
    {FFROM_TOKEN,       FOR_COMPOSE},
    {FADDRESS_TOKEN,    FOR_COMPOSE},
    {NULL,              FOR_NOTHING}
};


typedef struct {
        char         *name;
        char*        (*execute)();
        int          what_for;
} RULE_FCN;

#define INDEX_FCN       "_INDEX_"
#define REPLACE_FCN     "_REPLACE_"
#define REPLYSTR_FCN    "_RESTR_"
#define REPLY_FCN       "_REPLY_"
#define RESUB_FCN       "_RESUB_"
#define REXTRIM_FCN	"_REXTRIM_"
#define SAVE_FCN        "_SAVE_"
#define SIGNATURE_FCN   "_SIGNATURE_"
#define SMTP_FCN        "_SMTP_"
#define SORT_FCN        "_SORT_"
#define STARTUP_FCN     "_STARTUP_"
#define THRDSTYLE_FCN   "_THREADSTYLE_"
#define THRDINDEX_FCN   "_THREADINDEX_"
#define TRIM_FCN        "_TRIM_"


RULE_FCN rule_fcns[] = {
{SAVE_FCN,      extended_value, FOR_SAVE},
{REPLY_FCN,     extended_value, FOR_REPLY_INTRO},
{TRIM_FCN,      trim,		FOR_TRIM | FOR_RESUB},
{REPLACE_FCN,   extended_value,	FOR_REPLACE},
{SORT_FCN,      raw_value,	FOR_SORT},
{INDEX_FCN,     raw_value,	FOR_INDEX},
{REPLYSTR_FCN,  raw_value,	FOR_COMPOSE},
{SIGNATURE_FCN, raw_value,	FOR_COMPOSE},
{RESUB_FCN,     extended_value,	FOR_RESUB},
{STARTUP_FCN,   raw_value,	FOR_STARTUP},
{REXTRIM_FCN,   rextrim,	FOR_TRIM | FOR_RESUB},
{THRDSTYLE_FCN, raw_value,	FOR_THREAD},
{THRDINDEX_FCN, raw_value,	FOR_THREAD},
{SMTP_FCN,      raw_value,	FOR_COMPOSE}
};

#define NFCN    (sizeof(rule_fcns)/sizeof(rule_fcns[0]))

#define STRICT  0x1
#define RELAXED 0x2

